"""
Adobe DNG SDK Conversion Process
================================

Defines various objects implementing raw conversion based on *Adobe DNG SDK*
and *dcraw*:

-   :func:`colour_hdri.convert_raw_files_to_dng_files`
-   :func:`colour_hdri.convert_dng_files_to_intermediate_files`
-   :func:`colour_hdri.read_dng_files_exif_tags`
"""

from __future__ import annotations

import logging
import numpy as np
import os
import platform
import re
import shlex
import subprocess  # nosec

from colour.hints import (
    Callable,
    Boolean,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
)
from colour.utilities import CaseInsensitiveMapping, warning
from colour.utilities.documentation import (
    DocstringText,
    is_documentation_building,
)

from colour_hdri.utilities import (
    EXIFTag,
    parse_exif_array,
    parse_exif_number,
    parse_exif_string,
    path_exists,
    read_exif_tags,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "RAW_CONVERTER",
    "RAW_CONVERSION_ARGUMENTS",
    "RAW_D_CONVERSION_ARGUMENTS",
    "DNG_CONVERTER",
    "DNG_CONVERSION_ARGUMENTS",
    "DNG_EXIF_TAGS_BINDING",
    "convert_raw_files_to_dng_files",
    "convert_dng_files_to_intermediate_files",
    "read_dng_files_exif_tags",
]

_IS_MACOS_PLATFORM: Boolean = platform.system() == "Darwin"
"""Whether the current platform is *macOS*."""

_IS_WINDOWS_PLATFORM: Boolean = platform.system() in ("Windows", "Microsoft")
"""Whether the current platform is *Windows*."""

RAW_CONVERTER: str = "dcraw"
if is_documentation_building():  # pragma: no cover
    RAW_CONVERTER = DocstringText(RAW_CONVERTER)
    RAW_CONVERTER.__doc__ = """
Command line raw conversion application, usually Dave Coffin's *dcraw*.
"""

RAW_CONVERSION_ARGUMENTS: str = '-t 0 -D -W -4 -T "{0}"'
if _IS_WINDOWS_PLATFORM:
    RAW_CONVERSION_ARGUMENTS = RAW_CONVERSION_ARGUMENTS.replace('"', "")
if is_documentation_building():  # pragma: no cover
    RAW_CONVERSION_ARGUMENTS = DocstringText(RAW_CONVERSION_ARGUMENTS)
    RAW_CONVERSION_ARGUMENTS.__doc__ = """
Arguments for the command line raw conversion application for non
demosaiced linear *tiff* file format output.
"""

RAW_D_CONVERSION_ARGUMENTS: str = '-t 0 -H 1 -r 1 1 1 1 -4 -q 3 -o 0 -T "{0}"'
if _IS_WINDOWS_PLATFORM:
    RAW_D_CONVERSION_ARGUMENTS = RAW_D_CONVERSION_ARGUMENTS.replace('"', "")
if is_documentation_building():  # pragma: no cover
    RAW_D_CONVERSION_ARGUMENTS = DocstringText(RAW_D_CONVERSION_ARGUMENTS)
    RAW_D_CONVERSION_ARGUMENTS.__doc__ = """
Arguments for the command line raw conversion application for demosaiced
linear *tiff* file format output.
"""

if _IS_MACOS_PLATFORM:
    DNG_CONVERTER: Optional[str] = (
        "/Applications/Adobe DNG Converter.app/Contents/"
        "MacOS/Adobe DNG Converter"
    )
elif _IS_WINDOWS_PLATFORM:
    DNG_CONVERTER: Optional[  # type: ignore[no-redef]
        str
    ] = "Adobe DNG Converter"
else:
    warning('"Adobe DNG Converter" is not available on your platform!')
    DNG_CONVERTER: Optional[str] = None  # type: ignore[no-redef]

if DNG_CONVERTER is not None:
    if is_documentation_building():  # pragma: no cover
        DNG_CONVERTER = DocstringText(DNG_CONVERTER)
        DNG_CONVERTER.__doc__ = """
Command line *DNG* conversion application, usually *Adobe DNG Converter*.
"""

DNG_CONVERSION_ARGUMENTS: str = '-cr7.1 -l -d "{0}" "{1}"'
if _IS_WINDOWS_PLATFORM:
    DNG_CONVERSION_ARGUMENTS = DNG_CONVERSION_ARGUMENTS.replace('"', "")
if is_documentation_building():  # pragma: no cover
    DNG_CONVERSION_ARGUMENTS = DocstringText(DNG_CONVERSION_ARGUMENTS)
    DNG_CONVERSION_ARGUMENTS.__doc__ = """
Arguments for the command line *dng* conversion application.
"""

DNG_EXIF_TAGS_BINDING: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "EXIF": CaseInsensitiveMapping(
            {
                "Make": (parse_exif_string, None),
                "Camera Model Name": (parse_exif_string, None),
                "Camera Serial Number": (parse_exif_string, None),
                "Lens Model": (parse_exif_string, None),
                "DNG Lens Info": (parse_exif_string, None),
                "Focal Length": (parse_exif_number, None),
                "Exposure Time": (parse_exif_number, None),
                "F Number": (parse_exif_number, None),
                "ISO": (parse_exif_number, None),
                "CFA Pattern 2": (
                    lambda x: parse_exif_array(x, np.int_),
                    None,
                ),
                "CFA Plane Color": (
                    lambda x: parse_exif_array(x, np.int_),
                    None,
                ),
                "Black Level Repeat Dim": (
                    lambda x: parse_exif_array(x, np.int_),
                    None,
                ),
                "Black Level": (lambda x: parse_exif_array(x, np.int_), None),
                "White Level": (lambda x: parse_exif_array(x, np.int_), None),
                "Samples Per Pixel": (
                    lambda x: parse_exif_number(x, np.int_),
                    None,
                ),
                "Active Area": (lambda x: parse_exif_array(x, np.int_), None),
                "Orientation": (lambda x: parse_exif_number(x, np.int_), None),
                "Camera Calibration Sig": (parse_exif_string, None),
                "Profile Calibration Sig": (parse_exif_string, None),
                "Calibration Illuminant 1": (
                    lambda x: parse_exif_number(x, np.int_),
                    17,
                ),
                "Calibration Illuminant 2": (
                    lambda x: parse_exif_number(x, np.int_),
                    21,
                ),
                "Color Matrix 1": (
                    lambda x: parse_exif_array(x, np.float_, (3, 3)),
                    "1 0 0 0 1 0 0 0 1",
                ),
                "Color Matrix 2": (
                    lambda x: parse_exif_array(x, np.float_, (3, 3)),
                    "1 0 0 0 1 0 0 0 1",
                ),
                "Camera Calibration 1": (
                    lambda x: parse_exif_array(x, np.float_, (3, 3)),
                    "1 0 0 0 1 0 0 0 1",
                ),
                "Camera Calibration 2": (
                    lambda x: parse_exif_array(x, np.float_, (3, 3)),
                    "1 0 0 0 1 0 0 0 1",
                ),
                "Analog Balance": (
                    lambda x: parse_exif_array(x, np.float_),
                    "1 1 1",
                ),
                "Reduction Matrix 1": (
                    lambda x: parse_exif_array(x, np.float_, (3, 3)),
                    "1 0 0 0 1 0 0 0 1",
                ),
                "Reduction Matrix 2": (
                    lambda x: parse_exif_array(x, np.float_, (3, 3)),
                    "1 0 0 0 1 0 0 0 1",
                ),
                "Forward Matrix 1": (
                    lambda x: parse_exif_array(x, np.float_, (3, 3)),
                    "1 0 0 0 1 0 0 0 1",
                ),
                "Forward Matrix 2": (
                    lambda x: parse_exif_array(x, np.float_, (3, 3)),
                    "1 0 0 0 1 0 0 0 1",
                ),
                "As Shot Neutral": (
                    lambda x: parse_exif_array(x, np.float_),
                    "1 1 1",
                ),
                "Baseline Exposure": (
                    lambda x: parse_exif_number(x, np.float_),
                    None,
                ),
                "Baseline Noise": (
                    lambda x: parse_exif_number(x, np.float_),
                    None,
                ),
            }
        )
    }
)
DNG_EXIF_TAGS_BINDING.__doc__ = """
Exif tags binding for a *dng* file.
"""


def convert_raw_files_to_dng_files(
    raw_files: Sequence[str], output_directory: str
) -> List[str]:
    """
    Convert given raw files to *dng* files using given output directory.

    Parameters
    ----------
    raw_files
        Raw files to convert to *dng* files.
    output_directory
        Output directory.

    Returns
    -------
    :class:`list`
        *dng* files.

    Raises
    ------
    RuntimeError
        If the *Adobe DNG Converter* is not available.
    """

    if DNG_CONVERTER is not None:
        dng_files = []
        for raw_file in raw_files:
            raw_file_extension = os.path.splitext(raw_file)[1]
            dng_file = os.path.join(
                output_directory,
                os.path.basename(
                    re.sub(f"{raw_file_extension}$", ".dng", raw_file)
                ),
            )

            if path_exists(dng_file):
                os.remove(dng_file)

            logging.info(f'Converting "{raw_file}" file to "{dng_file}" file.')

            command = [DNG_CONVERTER] + shlex.split(
                DNG_CONVERSION_ARGUMENTS.format(output_directory, raw_file),
                posix=not _IS_WINDOWS_PLATFORM,
            )

            subprocess.call(command, shell=_IS_WINDOWS_PLATFORM)  # nosec

            dng_files.append(dng_file)

        return dng_files
    else:
        raise RuntimeError('The "Adobe DNG Converter" is not available!')


def convert_dng_files_to_intermediate_files(
    dng_files: Sequence[str],
    output_directory: str,
    demosaicing: Boolean = False,
) -> List[str]:
    """
    Convert given *dng* files to intermediate *tiff* files using given output
    directory.

    Parameters
    ----------
    dng_files
        *dng* files to convert to intermediate *tiff* files.
    output_directory
        Output directory.
    demosaicing
        Perform demosaicing on conversion.

    Returns
    -------
    :class:`list`
        Intermediate *tiff* files.
    """

    intermediate_files = []
    for dng_file in dng_files:
        intermediate_file = re.sub("\\.dng$", ".tiff", dng_file)

        if path_exists(intermediate_file):
            os.remove(intermediate_file)

        logging.info(
            f'Converting "{dng_file}" file to "{intermediate_file}" file.'
        )

        raw_conversion_arguments = (
            RAW_D_CONVERSION_ARGUMENTS
            if demosaicing
            else RAW_CONVERSION_ARGUMENTS
        )
        command = [RAW_CONVERTER] + shlex.split(
            raw_conversion_arguments.format(dng_file),
            posix=not _IS_WINDOWS_PLATFORM,
        )

        subprocess.call(command, shell=_IS_WINDOWS_PLATFORM)  # nosec

        tiff_file = os.path.join(
            output_directory, os.path.basename(intermediate_file)
        )
        if tiff_file != intermediate_file:
            if path_exists(tiff_file):
                os.remove(tiff_file)

            os.rename(intermediate_file, tiff_file)

        intermediate_files.append(tiff_file)

    return intermediate_files


def read_dng_files_exif_tags(
    dng_files: Sequence[str],
    exif_tags_binding: Mapping[
        str, Mapping[str, Tuple[Callable, Optional[str]]]
    ] = DNG_EXIF_TAGS_BINDING,
) -> List[CaseInsensitiveMapping]:
    """
    Read given *dng* files exif tags using given binding.

    Parameters
    ----------
    dng_files
        *dng* files to read the exif tags from.
    exif_tags_binding
        Exif tags binding.

    Returns
    -------
    :class:`list`
        *dng* files exif tags.
    """

    dng_files_exif_tags = []
    for dng_file in dng_files:
        exif_tags = read_exif_tags(dng_file)
        binding = CaseInsensitiveMapping()
        for group, tags in exif_tags_binding.items():
            binding[group] = CaseInsensitiveMapping()
            for tag, (parser, default) in tags.items():
                exif_tag = exif_tags[group].get(tag)
                if exif_tag is None:
                    binding[group][tag] = (
                        default
                        if default is None
                        else parser(EXIFTag(value=default))
                    )
                else:
                    binding[group][tag] = parser(exif_tag[0])

        dng_files_exif_tags.append(binding)

    return dng_files_exif_tags



"""
RGB Colourspace & Transformations
=================================

Defines the following *RGB* colourspace transformations:

-   :func:`colour_hdri.camera_space_to_RGB`
-   :func:`colour_hdri.camera_space_to_sRGB`
"""

from __future__ import annotations

import numpy as np

from colour.algebra import matrix_dot, vector_dot
from colour.hints import ArrayLike, NDArray
from colour.models import RGB_COLOURSPACES

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "camera_space_to_RGB",
    "camera_space_to_sRGB",
]


def camera_space_to_RGB(
    RGB: ArrayLike,
    M_XYZ_to_camera_space: ArrayLike,
    matrix_RGB_to_XYZ: ArrayLike,
) -> NDArray:
    """
    Convert given *RGB* array from *camera space* to given *RGB* colourspace.

    Parameters
    ----------
    RGB
        Camera space *RGB* colourspace array.
    M_XYZ_to_camera_space
        Matrix converting from *CIE XYZ* tristimulus values to *camera space*.
    matrix_RGB_to_XYZ
        Matrix converting from *RGB* colourspace to *CIE XYZ* tristimulus
        values.

    Returns
    -------
    :class:`numpy.ndarray`
        *RGB* colourspace array.

    Examples
    --------
    >>> RGB = np.array([0.80660, 0.81638, 0.65885])
    >>> M_XYZ_to_camera_space = np.array([
    ...     [0.47160000, 0.06030000, -0.08300000],
    ...     [-0.77980000, 1.54740000, 0.24800000],
    ...     [-0.14960000, 0.19370000, 0.66510000]])
    >>> matrix_RGB_to_XYZ = np.array([
    ...     [0.41238656, 0.35759149, 0.18045049],
    ...     [0.21263682, 0.71518298, 0.07218020],
    ...     [0.01933062, 0.11919716, 0.95037259]])
    >>> camera_space_to_RGB(
    ...     RGB,
    ...     M_XYZ_to_camera_space,
    ...     matrix_RGB_to_XYZ)  # doctest: +ELLIPSIS
    array([ 0.7564180...,  0.8683192...,  0.6044589...])
    """

    M_RGB_camera = matrix_dot(M_XYZ_to_camera_space, matrix_RGB_to_XYZ)

    M_RGB_camera /= np.transpose(np.sum(M_RGB_camera, axis=1)[np.newaxis])

    RGB_f = vector_dot(np.linalg.inv(M_RGB_camera), RGB)

    return RGB_f


def camera_space_to_sRGB(
    RGB: ArrayLike, M_XYZ_to_camera_space: ArrayLike
) -> NDArray:
    """
    Convert given *RGB* array from *camera space* to *sRGB* colourspace.

    Parameters
    ----------
    RGB
        Camera space *RGB* colourspace array.
    M_XYZ_to_camera_space
        Matrix converting from *CIE XYZ* tristimulus values to *camera space*.

    Returns
    -------
    :class:`numpy.ndarray`
        *sRGB* colourspace array.

    Examples
    --------
    >>> RGB = np.array([0.80660, 0.81638, 0.65885])
    >>> M_XYZ_to_camera_space = np.array([
    ...     [0.47160000, 0.06030000, -0.08300000],
    ...     [-0.77980000, 1.54740000, 0.24800000],
    ...     [-0.14960000, 0.19370000, 0.66510000]])
    >>> camera_space_to_sRGB(RGB, M_XYZ_to_camera_space)  # doctest: +ELLIPSIS
    array([ 0.7564350...,  0.8683155...,  0.6044706...])
    """

    return camera_space_to_RGB(
        RGB, M_XYZ_to_camera_space, RGB_COLOURSPACES["sRGB"].matrix_RGB_to_XYZ
    )



"""
Weighting Functions
===================

Defines the weighting function objects used when generating radiance images:

-   :func:`colour_hdri.normal_distribution_function`
-   :func:`colour_hdri.hat_function`
-   :func:`colour_hdri.weighting_function_Debevec1997`

References
----------
-   :cite:`Debevec1997a` : Debevec, P. E., & Malik, J. (1997). Recovering high
    dynamic range radiance maps from photographs. Proceedings of the 24th
    Annual Conference on Computer Graphics and Interactive Techniques -
    SIGGRAPH "97, August, 369-378. doi:10.1145/258734.258884
"""

from __future__ import annotations

import numpy as np

from colour.hints import ArrayLike, Floating, NDArray

from colour.utilities import as_float_array

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "normal_distribution_function",
    "hat_function",
    "weighting_function_Debevec1997",
]


def normal_distribution_function(
    a: ArrayLike, mu: Floating = 0.5, sigma: Floating = 0.15
) -> NDArray:
    """
    Return given array weighted by a normal distribution function.

    Parameters
    ----------
    a
        Array to apply the weighting function onto.
    mu
        Mean or expectation.
    sigma
        Standard deviation.

    Returns
    -------
    :class:`numpy.ndarray`
        Weighted array.

    Examples
    --------
    >>> normal_distribution_function(np.linspace(0, 1, 10))
    array([ 0.00386592,  0.03470859,  0.18002174,  0.53940751,  0.93371212,
            0.93371212,  0.53940751,  0.18002174,  0.03470859,  0.00386592])
    """

    a = as_float_array(a)

    return np.exp(-np.power(a - mu, 2) / (2 * np.power(sigma, 2)))


def hat_function(a: ArrayLike) -> NDArray:
    """
    Return given array weighted by a hat function.

    Parameters
    ----------
    a
        Array to apply the weighting function onto.

    Returns
    -------
    :class:`numpy.ndarray`
        Weighted array.

    Examples
    --------
    >>> hat_function(np.linspace(0, 1, 10))
    array([ 0.        ,  0.95099207,  0.99913557,  0.99999812,  1.        ,
            1.        ,  0.99999812,  0.99913557,  0.95099207,  0.        ])
    """

    a = as_float_array(a)

    return 1 - (2 * a - 1) ** 12


def weighting_function_Debevec1997(
    a: ArrayLike, domain_l: Floating = 0.01, domain_h: Floating = 0.99
) -> NDArray:
    """
    Return given array weighted by *Debevec (1997)* function.

    Parameters
    ----------
    a
        Array to apply the weighting function onto.
    domain_l
        Domain lowest possible value, values less than ``domain_l`` will be set
        to zero.
    domain_h
        Domain highest possible value, values greater than ``domain_h`` will be
        set to zero.

    Returns
    -------
    :class:`numpy.ndarray`
        Weighted array.

    References
    ----------
    :cite:`Debevec1997a`

    Examples
    --------
    >>> weighting_function_Debevec1997(np.linspace(0, 1, 10))
    array([ 0.        ,  0.23273657,  0.48849105,  0.74424552,  1.        ,
            1.        ,  0.74424552,  0.48849105,  0.23273657,  0.        ])
    """

    a = as_float_array(a)

    w = np.zeros(a.shape)

    mask = np.where(a <= (domain_l + domain_h) / 2, True, False)
    w[mask] = a[mask] - domain_l
    w[~mask] = domain_h - a[~mask]
    w /= np.max(w)

    w[w < 0] = 0

    return w



"""
Debevec (1997) Camera Response Function Computation
===================================================

Defines the *Debevec (1997)* camera responses computation objects:

-   :func:`colour_hdri.g_solve`
-   :func:`colour_hdri.camera_response_functions_Debevec1997`

See Also
--------
`Colour - HDRI - Examples: Merge from Low Dynamic Range Files Jupyter Notebook
<https://github.com/colour-science/colour-hdri/\
blob/master/colour_hdri/examples/examples_merge_from_ldr_files.ipynb>`__

References
----------
-   :cite:`Debevec1997a` : Debevec, P. E., & Malik, J. (1997). Recovering high
    dynamic range radiance maps from photographs. Proceedings of the 24th
    Annual Conference on Computer Graphics and Interactive Techniques -
    SIGGRAPH "97, August, 369-378. doi:10.1145/258734.258884
"""

from __future__ import annotations

import numpy as np
from functools import partial

from colour.hints import (
    Any,
    ArrayLike,
    Boolean,
    Callable,
    Dict,
    Floating,
    Integer,
    NDArray,
    Optional,
    Tuple,
)
from colour.utilities import as_float_array, as_int_array, tstack

from colour_hdri.exposure import average_luminance
from colour_hdri.generation import weighting_function_Debevec1997
from colour_hdri.sampling import samples_Grossberg2003
from colour_hdri.utilities import ImageStack

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "g_solve",
    "extrapolating_function_polynomial",
    "camera_response_functions_Debevec1997",
]


def g_solve(
    Z: ArrayLike,
    B: ArrayLike,
    l_s: Floating = 30,
    w: Callable = weighting_function_Debevec1997,
    n: Integer = 256,
) -> Tuple[NDArray, NDArray]:
    """
    Given a set of pixel values observed for several pixels in several images
    with different exposure times, this function returns the imaging system's
    response function :math:`g` as well as the log film irradiance values
    :math:`lE` for the observed pixels.

    Parameters
    ----------
    Z
        Set of pixel values observed for several pixels in several images.
    B
        Log :math:`\\Delta t`, or log shutter speed for images.
    l_s
        :math:`\\lambda` smoothing term.
    w
        Weighting function :math:`w`.
    n
        :math:`n` constant.

    Returns
    -------
    :class:`tuple`
        Camera response functions :math:`g(z)` and log film irradiance values
        :math:`lE`.

    References
    ----------
    :cite:`Debevec1997a`
    """

    Z = as_int_array(Z)
    B = as_float_array(B)

    Z_x, Z_y = Z.shape

    A = np.zeros((Z_x * Z_y + n + 1, n + Z_x))
    b = np.zeros((A.shape[0], 1))
    w_n = w(np.linspace(0, 1, n))

    k = 0
    for i in np.arange(Z_x):
        for j in np.arange(Z_y):
            Z_c = Z[i, j]
            w_ij = w_n[Z_c]
            A[k, Z_c] = w_ij
            A[k, n + i] = -w_ij
            b[k] = w_ij * B[j]
            k += 1

    A[k, np.int_(n / 2)] = 1
    k += 1

    for i in np.arange(1, n - 1, 1):
        A[k, i - 1] = l_s * w_n[i]
        A[k, i + 0] = l_s * w_n[i] * -2
        A[k, i + 1] = l_s * w_n[i]
        k += 1

    x = np.squeeze(np.linalg.lstsq(A, b, rcond=None)[0])

    g = x[0:n]
    lE = x[n : x.shape[0]]

    return g, lE


def extrapolating_function_polynomial(
    crfs: ArrayLike,
    weighting_function: Callable,
    degree: Integer = 7,
    **kwargs: Any,
) -> NDArray:
    """
    Polynomial extrapolating function used to handle zero-weighted data of
    given camera response functions.

    The extrapolation occurs where the weighting function masks fully the
    camera response functions, e.g. at both ends for *Debevec (1997)*.

    Parameters
    ----------
    crfs
        Camera response functions :math:`g(z)`.
    weighting_function
        Weighting function :math:`w`.
    degree
        Degree of the extrapolating function polynomial.

    Other Parameters
    ----------------
    kwargs
        Keyword arguments.

    Returns
    -------
    :class:`numpy.ndarray`
        Extrapolated camera response functions :math:`g(z)`.
    """

    crfs = as_float_array(crfs)

    samples = np.linspace(0, 1, crfs.shape[0])
    mask = ~(weighting_function(samples) == 0)

    for x in range(crfs.shape[-1]):
        coefficients = np.polyfit(samples[mask], crfs[mask, x], degree)
        polynomial = np.poly1d(coefficients)
        crfs[~mask, x] = polynomial(samples[~mask])

    return crfs


def camera_response_functions_Debevec1997(
    image_stack: ImageStack,
    sampling_function: Callable = samples_Grossberg2003,
    sampling_function_kwargs: Optional[Dict] = None,
    weighting_function: Callable = weighting_function_Debevec1997,
    weighting_function_kwargs: Optional[Dict] = None,
    extrapolating_function: Callable = extrapolating_function_polynomial,
    extrapolating_function_kwargs: Optional[Dict] = None,
    l_s: Floating = 30,
    n: Integer = 256,
    normalise: Boolean = True,
) -> NDArray:
    """
    Return the camera response functions for given image stack using
    *Debevec (1997)* method.

    Image channels are sampled with :math:`s` sampling function and the output
    samples are passed to :func:`colour_hdri.g_solve`.

    Parameters
    ----------
    image_stack
        Stack of single channel or multi-channel floating point images.
    sampling_function
        Sampling function :math:`s`.
    sampling_function_kwargs
        Arguments to use when calling the sampling function.
    weighting_function
        Weighting function :math:`w`.
    weighting_function_kwargs
        Arguments to use when calling the weighting function.
    extrapolating_function
        Extrapolating function used to handle zero-weighted data.
    extrapolating_function_kwargs
        Arguments to use when calling the extrapolating function.
    l_s
        :math:`\\lambda` smoothing term.
    n
        :math:`n` constant.
    normalise
        Enables the camera response functions normalisation.

    Returns
    -------
    :class:`numpy.ndarray`
        Camera response functions :math:`g(z)`.

    References
    ----------
    :cite:`Debevec1997a`
    """

    if sampling_function_kwargs is None:
        sampling_function_kwargs = {}

    if weighting_function_kwargs is None:
        weighting_function_kwargs = {}

    if extrapolating_function_kwargs is None:
        extrapolating_function_kwargs = {}

    s_o = sampling_function(image_stack.data, **sampling_function_kwargs)

    L_l = np.log(
        1
        / average_luminance(
            image_stack.f_number, image_stack.exposure_time, image_stack.iso
        )
    )

    w = partial(weighting_function, **weighting_function_kwargs)

    g_c = [
        g_solve(s_o[..., x], L_l, l_s, w, n)[0] for x in range(s_o.shape[-1])
    ]
    crfs = np.exp(tstack(np.array(g_c)))

    if extrapolating_function is not None:
        crfs = extrapolating_function(crfs, w, **extrapolating_function_kwargs)

    if normalise:
        crfs /= np.max(crfs, axis=0)

    return crfs



from .common import vivification, vivified_to_dict, path_exists, filter_files
from .exif import (
    EXIF_EXECUTABLE,
    EXIFTag,
    parse_exif_string,
    parse_exif_number,
    parse_exif_fraction,
    parse_exif_array,
    parse_exif_data,
    read_exif_tags,
    copy_exif_tags,
    update_exif_tags,
    delete_exif_tags,
    read_exif_tag,
    write_exif_tag,
)
from .image import Metadata, Image, ImageStack

__all__ = []
__all__ += [
    "vivification",
    "vivified_to_dict",
    "path_exists",
    "filter_files",
]
__all__ += [
    "EXIF_EXECUTABLE",
    "EXIFTag",
    "parse_exif_string",
    "parse_exif_number",
    "parse_exif_fraction",
    "parse_exif_array",
    "parse_exif_data",
    "read_exif_tags",
    "copy_exif_tags",
    "update_exif_tags",
    "delete_exif_tags",
    "read_exif_tag",
    "write_exif_tag",
]
__all__ += [
    "Metadata",
    "Image",
    "ImageStack",
]



"""
HDRI / Radiance Image Plotting
==============================

Defines the HDRI / radiance image plotting objects:

-   :func:`colour_hdri.plotting.plot_radiance_image_strip`
"""

from __future__ import annotations

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

from colour.hints import Any, ArrayLike, Callable, Floating, Integer, Tuple
from colour.plotting import CONSTANTS_COLOUR_STYLE, override_style, render
from colour.utilities import as_float_array

from colour_hdri.exposure import adjust_exposure

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "plot_radiance_image_strip",
]


@override_style()
def plot_radiance_image_strip(
    image: ArrayLike,
    count: Integer = 5,
    ev_steps: Floating = -2,
    cctf_encoding: Callable = CONSTANTS_COLOUR_STYLE.colour.colourspace.cctf_encoding,
    **kwargs: Any,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Plot given HDRI / radiance image as strip of images of varying exposure.

    Parameters
    ----------
    image
         HDRI / radiance image to plot.
    count
        Strip images count.
    ev_steps
        Exposure variation for each image of the strip.
    cctf_encoding
        Encoding colour component transfer function / opto-electronic
        transfer function used for plotting.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.display`},
        Please refer to the documentation of the previously listed definition.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.
    """

    image = as_float_array(image)

    grid = matplotlib.gridspec.GridSpec(1, count)
    grid.update(wspace=0, hspace=0)

    height, width, _channel = image.shape
    for i in range(count):
        ev = i * ev_steps
        axis = plt.subplot(grid[i])
        axis.imshow(np.clip(cctf_encoding(adjust_exposure(image, ev)), 0, 1))
        axis.text(
            width * 0.05, height - height * 0.05, f"EV {ev}", color=(1, 1, 1)
        )
        axis.set_xticks([])
        axis.set_yticks([])
        axis.set_aspect("equal")

    return render(**kwargs)



# !/usr/bin/env python
"""
Defines the unit tests for the
:mod:`colour_hdri.tonemapping.global_operators.operators` module.
"""

import numpy as np
import unittest

from colour_hdri.tonemapping import (
    tonemapping_operator_simple,
    tonemapping_operator_normalisation,
    tonemapping_operator_gamma,
    tonemapping_operator_logarithmic,
    tonemapping_operator_exponential,
    tonemapping_operator_logarithmic_mapping,
    tonemapping_operator_exponentiation_mapping,
    tonemapping_operator_Schlick1994,
    tonemapping_operator_Tumblin1999,
    tonemapping_operator_Reinhard2004,
    tonemapping_operator_filmic,
)
from colour_hdri.tonemapping.global_operators.operators import log_average

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestLogAverage",
    "TestTonemappingOperatorSimple",
    "TestTonemappingOperatorNormalisation",
    "TestTonemappingOperatorGamma",
    "TestTonemappingOperatorLogarithmic",
    "TestTonemappingOperatorExponential",
    "TestTonemappingOperatorLogarithmicMapping",
    "TestTonemappingOperatorExponentiationMapping",
    "TestTonemappingOperatorSchlick1994",
    "TestTonemappingOperatorTumblin1999",
    "TestTonemappingOperatorReinhard2004",
    "TestTonemappingOperatorFilmic",
]


class TestLogAverage(unittest.TestCase):
    """
    Define :func:`colour_hdri.utilities.common.log_average` definition unit
    tests methods.
    """

    def test_log_average(self):
        """Test :func:`colour_hdri.utilities.common.log_average` definition."""

        np.testing.assert_allclose(
            log_average(np.linspace(0, 10, 10)),
            np.array(0.125071409675722),
            rtol=0.025,
            atol=0.025,
        )


class TestTonemappingOperatorSimple(unittest.TestCase):
    """
    Define :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_simple` definition unit tests methods.
    """

    def test_tonemapping_operator_simple(self):
        """
        Test :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_simple` definition.
        """

        np.testing.assert_almost_equal(
            tonemapping_operator_simple(
                np.array(
                    [
                        [
                            [0.48046875, 0.35156256, 0.23632812],
                            [1.39843753, 0.55468757, 0.39062594],
                        ],
                        [
                            [4.40625388, 2.15625895, 1.34375372],
                            [6.59375023, 3.43751395, 2.21875829],
                        ],
                    ]
                )
            ),
            np.array(
                [
                    [
                        [0.32453826, 0.26011564, 0.19115324],
                        [0.58306189, 0.35678395, 0.28089936],
                    ],
                    [
                        [0.81502903, 0.68316922, 0.57333401],
                        [0.86831276, 0.7746486, 0.68932119],
                    ],
                ]
            ),
            decimal=7,
        )


class TestTonemappingOperatorNormalisation(unittest.TestCase):
    """
    Define :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_normalisation` definition unit tests methods.
    """

    def test_tonemapping_operator_normalisation(self):
        """
        Test :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_normalisation` definition.
        """

        np.testing.assert_almost_equal(
            tonemapping_operator_normalisation(
                np.array(
                    [
                        [
                            [0.48046875, 0.35156256, 0.23632812],
                            [1.39843753, 0.55468757, 0.39062594],
                        ],
                        [
                            [4.40625388, 2.15625895, 1.34375372],
                            [6.59375023, 3.43751395, 2.21875829],
                        ],
                    ]
                )
            ),
            np.array(
                [
                    [
                        [0.11949973, 0.08743884, 0.05877832],
                        [0.34781223, 0.13795905, 0.09715449],
                    ],
                    [
                        [1.09590092, 0.53629369, 0.33421155],
                        [1.63996382, 0.85496088, 0.55183821],
                    ],
                ]
            ),
            decimal=7,
        )


class TestTonemappingOperatorGamma(unittest.TestCase):
    """
    Define :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_gamma` definition unit tests methods.
    """

    def test_tonemapping_operator_gamma(self):
        """
        Test :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_gamma` definition.
        """

        np.testing.assert_almost_equal(
            tonemapping_operator_gamma(
                np.array(
                    [
                        [
                            [0.48046875, 0.35156256, 0.23632812],
                            [1.39843753, 0.55468757, 0.39062594],
                        ],
                        [
                            [4.40625388, 2.15625895, 1.34375372],
                            [6.59375023, 3.43751395, 2.21875829],
                        ],
                    ]
                ),
                0.5,
                -5,
            ),
            np.array(
                [
                    [
                        [2.25439668e-04, 1.20699447e-04, 5.45419730e-05],
                        [1.90979251e-03, 3.00467090e-04, 1.49012329e-04],
                    ],
                    [
                        [1.89600325e-02, 4.54048111e-03, 1.76335357e-03],
                        [4.24585372e-02, 1.15395529e-02, 4.80750815e-03],
                    ],
                ]
            ),
            decimal=7,
        )


class TestTonemappingOperatorLogarithmic(unittest.TestCase):
    """
    Define :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_logarithmic` definition unit tests methods.
    """

    def test_tonemapping_operator_logarithmic(self):
        """
        Test :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_logarithmic` definition.
        """

        np.testing.assert_almost_equal(
            tonemapping_operator_logarithmic(
                np.array(
                    [
                        [
                            [0.48046875, 0.35156256, 0.23632812],
                            [1.39843753, 0.55468757, 0.39062594],
                        ],
                        [
                            [4.40625388, 2.15625895, 1.34375372],
                            [6.59375023, 3.43751395, 2.21875829],
                        ],
                    ]
                ),
                10,
                25,
            ),
            np.array(
                [
                    [
                        [0.43458200, 0.31798688, 0.21375781],
                        [0.88293445, 0.35021426, 0.24663032],
                    ],
                    [
                        [1.21686841, 0.59549075, 0.37110242],
                        [1.31992973, 0.68811779, 0.44414861],
                    ],
                ]
            ),
            decimal=7,
        )


class TestTonemappingOperatorExponential(unittest.TestCase):
    """
    Define :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_exponential` definition unit tests methods.
    """

    def test_tonemapping_operator_exponential(self):
        """
        Test :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_exponential` definition.
        """

        np.testing.assert_almost_equal(
            tonemapping_operator_exponential(
                np.array(
                    [
                        [
                            [0.48046875, 0.35156256, 0.23632812],
                            [1.39843753, 0.55468757, 0.39062594],
                        ],
                        [
                            [4.40625388, 2.15625895, 1.34375372],
                            [6.59375023, 3.43751395, 2.21875829],
                        ],
                    ]
                ),
                10,
                25,
            ),
            np.array(
                [
                    [
                        [0.14069770, 0.10294955, 0.06920496],
                        [0.38838808, 0.15405339, 0.10848855],
                    ],
                    [
                        [0.94076970, 0.46037817, 0.28690194],
                        [1.16837509, 0.60910795, 0.39315137],
                    ],
                ]
            ),
            decimal=7,
        )


class TestTonemappingOperatorLogarithmicMapping(unittest.TestCase):
    """
    Define :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_logarithmic_mapping` definition unit tests methods.
    """

    def test_tonemapping_operator_logarithmic_mapping(self):
        """
        Test :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_logarithmic_mapping` definition.
        """

        np.testing.assert_almost_equal(
            tonemapping_operator_logarithmic_mapping(
                np.array(
                    [
                        [
                            [0.48046875, 0.35156256, 0.23632812],
                            [1.39843753, 0.55468757, 0.39062594],
                        ],
                        [
                            [4.40625388, 2.15625895, 1.34375372],
                            [6.59375023, 3.43751395, 2.21875829],
                        ],
                    ]
                ),
                10,
                2.0,
            ),
            np.array(
                [
                    [
                        [0.83661533, 0.61215767, 0.41150591],
                        [1.45740883, 0.57807842, 0.40709841],
                    ],
                    [
                        [1.60813343, 0.78696148, 0.49042459],
                        [1.63996382, 0.85496088, 0.55183821],
                    ],
                ]
            ),
            decimal=7,
        )


class TestTonemappingOperatorExponentiationMapping(unittest.TestCase):
    """
    Define :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_exponentiation_mapping` definition unit tests methods.
    """

    def test_tonemapping_operator_exponentiation_mapping(self):
        """
        Test :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_exponentiation_mapping` definition.
        """

        np.testing.assert_almost_equal(
            tonemapping_operator_exponentiation_mapping(
                np.array(
                    [
                        [
                            [0.48046875, 0.35156256, 0.23632812],
                            [1.39843753, 0.55468757, 0.39062594],
                        ],
                        [
                            [4.40625388, 2.15625895, 1.34375372],
                            [6.59375023, 3.43751395, 2.21875829],
                        ],
                    ]
                ),
                0.5,
                2.0,
            ),
            np.array(
                [
                    [
                        [0.71427273, 0.52263867, 0.35132926],
                        [1.26051891, 0.49998241, 0.35210109],
                    ],
                    [
                        [1.5303143, 0.74887966, 0.46669247],
                        [1.63996382, 0.85496088, 0.55183821],
                    ],
                ]
            ),
            decimal=7,
        )


class TestTonemappingOperatorSchlick1994(unittest.TestCase):
    """
    Define :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_Schlick1994` definition unit tests methods.
    """

    def test_tonemapping_operator_Schlick1994(self):
        """
        Test :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_Schlick1994` definition.
        """

        np.testing.assert_almost_equal(
            tonemapping_operator_Schlick1994(
                np.array(
                    [
                        [
                            [0.48046875, 0.35156256, 0.23632812],
                            [1.39843753, 0.55468757, 0.39062594],
                        ],
                        [
                            [4.40625388, 2.15625895, 1.34375372],
                            [6.59375023, 3.43751395, 2.21875829],
                        ],
                    ]
                ),
                10,
            ),
            np.array(
                [
                    [
                        [0.65311658, 0.47789026, 0.32124839],
                        [1.32918845, 0.52722005, 0.37128257],
                    ],
                    [
                        [1.61964741, 0.79259601, 0.49393596],
                        [1.63996382, 0.85496088, 0.55183821],
                    ],
                ]
            ),
            decimal=7,
        )


class TestTonemappingOperatorTumblin1999(unittest.TestCase):
    """
    Define :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_Tumblin1999` definition unit tests methods.
    """

    def test_tonemapping_operator_Tumblin1999(self):
        """
        Test :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_Tumblin1999` definition.
        """

        np.testing.assert_almost_equal(
            tonemapping_operator_Tumblin1999(
                np.array(
                    [
                        [
                            [0.48046875, 0.35156256, 0.23632812],
                            [1.39843753, 0.55468757, 0.39062594],
                        ],
                        [
                            [4.40625388, 2.15625895, 1.34375372],
                            [6.59375023, 3.43751395, 2.21875829],
                        ],
                    ]
                ),
                50,
                75,
                85,
            ),
            np.array(
                [
                    [
                        [0.11466395, 0.08390046, 0.05639975],
                        [0.28234899, 0.11199318, 0.07886862],
                    ],
                    [
                        [0.64682951, 0.31653462, 0.19726043],
                        [0.86574718, 0.45133921, 0.29131885],
                    ],
                ]
            ),
            decimal=7,
        )


class TestTonemappingOperatorReinhard2004(unittest.TestCase):
    """
    Define :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_Reinhard2004` definition unit tests methods.
    """

    def test_tonemapping_operator_Reinhard2004(self):
        """
        Test :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_Reinhard2004` definition.
        """

        np.testing.assert_almost_equal(
            tonemapping_operator_Reinhard2004(
                np.array(
                    [
                        [
                            [0.48046875, 0.35156256, 0.23632812],
                            [1.39843753, 0.55468757, 0.39062594],
                        ],
                        [
                            [4.40625388, 2.15625895, 1.34375372],
                            [6.59375023, 3.43751395, 2.21875829],
                        ],
                    ]
                ),
                -5,
                0.5,
                0.5,
                0.5,
            ),
            np.array(
                [
                    [
                        [0.03388145, 0.03028025, 0.02267653],
                        [0.08415536, 0.04335064, 0.03388967],
                    ],
                    [
                        [0.14417912, 0.09625765, 0.07082831],
                        [0.15459508, 0.11021455, 0.08524267],
                    ],
                ]
            ),
            decimal=7,
        )


class TestTonemappingOperatorFilmic(unittest.TestCase):
    """
    Define :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_filmic` definition unit tests methods.
    """

    def test_tonemapping_operator_filmic(self):
        """
        Test :func:`colour_hdri.tonemapping.global_operators.operators.\
tonemapping_operator_filmic` definition.
        """

        np.testing.assert_almost_equal(
            tonemapping_operator_filmic(
                np.array(
                    [
                        [
                            [0.48046875, 0.35156256, 0.23632812],
                            [1.39843753, 0.55468757, 0.39062594],
                        ],
                        [
                            [4.40625388, 2.15625895, 1.34375372],
                            [6.59375023, 3.43751395, 2.21875829],
                        ],
                    ]
                ),
                0.5,
                0.5,
                0.5,
                0.5,
                0.05,
                0.5,
                4,
                15,
            ),
            np.array(
                [
                    [
                        [0.77200408, 0.69565383, 0.58310785],
                        [0.93794714, 0.80274647, 0.72279964],
                    ],
                    [
                        [1.00563196, 0.97235243, 0.93403169],
                        [1.01635759, 0.99657186, 0.97416808],
                    ],
                ]
            ),
            decimal=7,
        )


if __name__ == "__main__":
    unittest.main()



"""
Grossberg (2003) Histogram Based Image Sampling
===============================================

Defines the *Grossberg (2003)* histogram based image sampling objects:

-   :func:`colour_hdri.samples_Grossberg2003`

References
----------
-   :cite:`Banterle2014a` : Banterle, F., & Benedetti, L. (2014). PICCANTE: An
    Open and Portable Library for HDR Imaging.
-   :cite:`Grossberg2003g` : Grossberg, M. D., & Nayar, S. K. (2003).
    Determining the camera response from images: What is knowable? IEEE
    Transactions on Pattern Analysis and Machine Intelligence, 25(11),
    1455-1467. doi:10.1109/TPAMI.2003.1240119
"""

from __future__ import annotations

import numpy as np

from colour.hints import ArrayLike, Integer, NDArray
from colour.utilities import as_float_array, tsplit, tstack

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "samples_Grossberg2003",
]


def samples_Grossberg2003(
    image_stack: ArrayLike, samples: Integer = 1000, n: Integer = 256
) -> NDArray:
    """
    Return the samples for given image stack intensity histograms using
    *Grossberg (2003)* method.

    Parameters
    ----------
    image_stack
        Stack of single channel or multi-channel floating point images.
    samples
        Samples count.
    n
        Histograms bins count.

    Returns
    -------
    :class:`numpy.ndarray`
        Intensity histograms samples.

    References
    ----------
    :cite:`Banterle2014a`, :cite:`Grossberg2003g`
    """

    image_stack = as_float_array(image_stack)

    if image_stack.ndim == 3:
        channels_c = 1
    else:
        channels_c = image_stack.shape[-2]

    cdf_i = []
    for image in tsplit(image_stack):
        histograms = tstack(
            [
                np.histogram(image[..., c], n, range=(0, 1))[0]
                for c in np.arange(channels_c)
            ]
        )
        cdf = np.cumsum(histograms, axis=0)
        cdf_i.append(cdf.astype(np.float_) / np.max(cdf, axis=0))

    samples_cdf_i = np.zeros((samples, len(cdf_i), channels_c))
    samples_u = np.linspace(0, 1, samples)
    for i in np.arange(samples):
        for j in np.arange(channels_c):
            for k, cdf in enumerate(cdf_i):
                samples_cdf_i[i, k, j] = np.argmin(
                    np.abs(cdf[:, j] - samples_u[i])
                )

    return samples_cdf_i



"""
Digital Still Camera Exposure
=============================

Defines various objects for modeling Digital Still Camera (DSC) exposure:

-   :func:`colour_hdri.focal_plane_exposure`
-   :func:`colour_hdri.arithmetic_mean_focal_plane_exposure`
-   :func:`colour_hdri.saturation_based_speed_focal_plane_exposure`
-   :func:`colour_hdri.exposure_index_values`
-   :func:`colour_hdri.exposure_value_100`
-   :func:`colour_hdri.photometric_exposure_scale_factor_Lagarde2014`

References
----------
-   :cite:`ISO2006` : ISO. (2006). INTERNATIONAL STANDARD ISO12232-2006 -
    Photography - Digital still cameras - Determination of exposure index, ISO
    speed ratings, standard output sensitivity, and recommended exposure index.
-   :cite:`Lagarde2014` : Lagarde, Sébastian, & de Rousiers, C. (2014). Moving
    Frostbite to Physically Based Rendering 3.0. Siggraph 2014, 119.
"""

from __future__ import annotations

import numpy as np

from colour.hints import FloatingOrArrayLike, FloatingOrNDArray

from colour.utilities import as_float, as_float_array
from colour_hdri.exposure import (
    average_luminance,
    luminance_to_exposure_value,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "q_factor",
    "focal_plane_exposure",
    "arithmetic_mean_focal_plane_exposure",
    "saturation_based_speed_focal_plane_exposure",
    "exposure_index_values",
    "exposure_value_100",
    "photometric_exposure_scale_factor_Lagarde2014",
]


def q_factor(
    T: FloatingOrArrayLike = 9 / 10,
    f_v: FloatingOrArrayLike = 98 / 100,
    theta: FloatingOrArrayLike = 10,
) -> FloatingOrNDArray:
    """
    Compute the :math:`q` factor modeling the total lens vignetting and
    transmission attenuation.

    Parameters
    ----------
    T
        Transmission factor of the lens :math:`T`.
    f_v
        Vignetting factor :math:`f_v`.
    theta
        Angle of image point off axis :math:`\\theta`.

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        :math:`q` factor.

    References
    ----------
    :cite:`ISO2006`

    Examples
    --------
    >>> q_factor()  # doctest: +ELLIPSIS
    0.6515748...
    """

    T = as_float_array(T)
    f_v = as_float_array(f_v)
    theta = as_float_array(theta)

    return as_float(np.pi / 4 * T * f_v * np.cos(np.radians(theta)) ** 4)


def focal_plane_exposure(
    L: FloatingOrArrayLike,
    A: FloatingOrArrayLike,
    t: FloatingOrArrayLike,
    F: FloatingOrArrayLike,
    i: FloatingOrArrayLike,
    H_f: FloatingOrArrayLike,
    T: FloatingOrArrayLike = 9 / 10,
    f_v: FloatingOrArrayLike = 98 / 100,
    theta: FloatingOrArrayLike = 10,
) -> FloatingOrNDArray:
    """
    Compute the focal plane exposure :math:`H` in lux-seconds (:math:`lx.s`).

    Parameters
    ----------
    L
        Scene luminance :math:`L`, expressed in :math:`cd/m^2`.
    A
        Lens *F-Number* :math:`A`.
    t
        *Exposure Time* :math:`t`, expressed in seconds.
    F
        Lens focal length :math:`F`, expressed in meters.
    i
        Image distance :math:`i`, expressed in meters.
    H_f
        Focal plane flare exposure :math:`H_f`, expressed in lux-seconds
        (:math:`lx.s`).
    T
        Transmission factor of the lens :math:`T`.
    f_v
        Vignetting factor :math:`f_v`.
    theta
        Angle of image point off axis :math:`\\theta`.

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        Focal plane exposure :math:`H` in lux-seconds (:math:`lx.s`).

    Notes
    -----
    -   Focal plane exposure is also named luminous exposure or photometric
        exposure and is time-integrated illuminance.
    -   Object distance :math:`o`, focal length :math:`F`, and image distance
        :math:`i` are related by the thin-lens equation:
        :math:`\\cfrac{1}{f}=\\cfrac{1}{o}+\\cfrac{1}{i}`
    -   This method ignores the *ISO* arithmetic speed :math:`S` and is not
        concerned with determining an appropriate minimum or maximum exposure
        level.

    References
    ----------
    :cite:`ISO2006`

    Examples
    --------
    >>> focal_plane_exposure(4000, 8, 1 / 250, 50 / 1000, 50 / 1000, 0.0015)
    ... # doctest: +ELLIPSIS
    0.1643937...
    """

    L = as_float_array(L)
    t = as_float_array(t)
    A = as_float_array(A)
    F = as_float_array(F)
    i = as_float_array(i)
    H_f = as_float_array(H_f)

    q = q_factor(T, f_v, theta)

    H = q * (L * t * F**2) / (A**2 * i**2) + H_f

    return as_float(H)


def arithmetic_mean_focal_plane_exposure(
    L_a: FloatingOrArrayLike, A: FloatingOrArrayLike, t: FloatingOrArrayLike
) -> FloatingOrNDArray:
    """
    Compute the arithmetic mean focal plane exposure :math:`H_a` for a camera
    focused on infinity, :math:`H_f << H`, :math:`T=9/10`,
    :math:`\\theta =10^{\\circ}` and :math:`f_v=98/100`.

    Parameters
    ----------
    L_a
        Arithmetic scene luminance :math:`L_a`, expressed in :math:`cd/m^2`.
    A
        Lens *F-Number* :math:`A`.
    t
        *Exposure Time* :math:`t`, expressed in seconds.

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        Focal plane exposure :math:`H_a`.

    Notes
    -----
    -   Focal plane exposure is also named luminous exposure or photometric
        exposure and is time-integrated illuminance.
    -   Object distance :math:`o`, focal length :math:`F`, and image distance
        :math:`i` are related by the thin-lens equation:
        :math:`\\cfrac{1}{f}=\\cfrac{1}{o}+\\cfrac{1}{i}`
    -   This method ignores the *ISO* arithmetic speed :math:`S` and is not
        concerned with determining an appropriate minimum or maximum exposure
        level.

    References
    ----------
    :cite:`ISO2006`

    Examples
    --------
    >>> arithmetic_mean_focal_plane_exposure(4000, 8, 1 / 250)
    ... # doctest: +ELLIPSIS
    0.1628937...
    """

    H_a = focal_plane_exposure(L_a, A, t, 1, 1, 0, 9 / 10, 98 / 100, 10)

    return H_a


def saturation_based_speed_focal_plane_exposure(
    L: FloatingOrArrayLike,
    A: FloatingOrArrayLike,
    t: FloatingOrArrayLike,
    S: FloatingOrArrayLike,
    F: FloatingOrArrayLike = 50 / 1000,
    i: FloatingOrArrayLike = 1 / (-1 / 5 + 1 / (50 / 1000)),
    H_f: FloatingOrArrayLike = 0,
    T: FloatingOrArrayLike = 9 / 10,
    f_v: FloatingOrArrayLike = 98 / 100,
    theta: FloatingOrArrayLike = 10,
) -> FloatingOrNDArray:
    """
    Compute the Saturation-Based Speed (SBS) focal plane exposure
    :math:`H_{SBS}` in lux-seconds (:math:`lx.s`).

    The model implemented by this definition is appropriate to simulate a
    physical camera in an offline or realtime renderer.

    Parameters
    ----------
    L
        Scene luminance :math:`L`, expressed in :math:`cd/m^2`.
    A
        Lens *F-Number* :math:`A`.
    t
        *Exposure Time* :math:`t`, expressed in seconds.
    S
        *ISO* arithmetic speed :math:`S`.
    F
        Lens focal length :math:`F`, expressed in meters.
    i
        Image distance :math:`i`, expressed in meters.
    H_f
        Focal plane flare exposure :math:`H_f`, expressed in lux-seconds
        (:math:`lx.s`).
    T
        Transmission factor of the lens :math:`T`.
    f_v
        Vignetting factor :math:`f_v`.
    theta
        Angle of image point off axis :math:`\\theta`.

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        Saturation-Based Speed focal plane exposure :math:`H_{SBS}` in
        lux-seconds (:math:`lx.s`).

    Notes
    -----
    -   Focal plane exposure is also named luminous exposure or photometric
        exposure and is time-integrated illuminance.
    -   Object distance :math:`o`, focal length :math:`F`, and image distance
        :math:`i` are related by the thin-lens equation:
        :math:`\\cfrac{1}{f}=\\cfrac{1}{o}+\\cfrac{1}{i}`
    -   The image distance default value is that of an object located at 5m and
        imaged with a 50mm lens.
    -   The saturation based speed, :math:`S_{sat}`, of an electronic still
        picture camera is defined as: :math:`S_{sat}=\\cfrac{78}{H_{sat}}`
        where :math:`H_{sat}` is the minimum focal plane exposure, expressed in
        lux-seconds (:math:`lx.s`), that produces the maximum valid (not
        clipped or bloomed) camera output signal. This provides :math:`1/2`
        "stop" of headroom (41% additional headroom) for specular highlights
        above the signal level that would be obtained from a theoretical 100%
        reflectance object in the scene, so that a theoretical 141% reflectance
        object in the scene would produce a focal plane exposure of
        :math:`H_{sat}`.
    -   The focal plane exposure :math:`H_{SBS}` computed by this definition is
        almost equal to that given by scene luminance :math:`L` scaled with
        the output of :func:`colour_hdri.\
photometric_exposure_scale_factor_Lagarde2014` definition.

    References
    ----------
    :cite:`ISO2006`

    Examples
    --------
    >>> saturation_based_speed_focal_plane_exposure(  # doctest: +ELLIPSIS
    ...     4000, 8, 1 / 250, 400, 50 / 1000, 50 / 1000, 0.0015)
    0.8430446...
    """

    S = as_float_array(S)

    H = focal_plane_exposure(L, A, t, F, i, H_f, T, f_v, theta)

    H_SBS = H * S / 78

    return as_float(H_SBS)


def exposure_index_values(H_a: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Compute the exposure index values :math:`I_{EI}` from given focal plane
    exposure :math:`H_a`.

    Parameters
    ----------
    H_a
        Focal plane exposure :math:`H_a`.

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        Exposure index values :math:`I_{EI}`.

    References
    ----------
    :cite:`ISO2006`

    Examples
    --------
    >>> exposure_index_values(0.1628937086212269)  # doctest: +ELLIPSIS
    61.3897251...
    """

    return as_float(10 / as_float_array(H_a))


def exposure_value_100(
    N: FloatingOrArrayLike, t: FloatingOrArrayLike, S: FloatingOrArrayLike
) -> FloatingOrNDArray:
    """
    Compute the exposure value :math:`EV100` from given relative aperture
    *F-Number* :math:`N`, *Exposure Time* :math:`t` and *ISO* arithmetic
    speed :math:`S`.

    Parameters
    ----------
    N
        Relative aperture *F-Number* :math:`N`.
    t
       *Exposure Time* :math:`t`.
    S
        *ISO* arithmetic speed :math:`S`.

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        Exposure value :math:`EV100`.

    References
    ----------
    :cite:`ISO2006`, :cite:`Lagarde2014`

    Notes
    -----
    -   The underlying implementation uses the
        :func:`colour_hdri.luminance_to_exposure_value` and
        :func:`colour_hdri.average_luminance` definitions with same fixed value
        for the *reflected light calibration constant* :math:`k` which cancels
        its scaling effect and produces a value equal to
        :math:`log_2(\\cfrac{N^2}{t}) - log_2(\\cfrac{S}{100})` as given in
        :cite:`Lagarde2014`.

    Examples
    --------
    >>> exposure_value_100(8, 1 / 250, 400)  # doctest: +ELLIPSIS
    11.9657842...
    """

    return luminance_to_exposure_value(average_luminance(N, t, S), 100)


def photometric_exposure_scale_factor_Lagarde2014(
    EV100: FloatingOrArrayLike,
    T: FloatingOrArrayLike = 9 / 10,
    f_v: FloatingOrArrayLike = 98 / 100,
    theta: FloatingOrArrayLike = 10,
) -> FloatingOrNDArray:
    """
    Convert the exposure value :math:`EV100` to photometric exposure scale
    factor using *Lagarde and de Rousiers (2014)* formulation derived from the
    *ISO 12232:2006* *Saturation Based Sensitivity* (SBS) recommendation.

    The model implemented by this definition is appropriate to simulate a
    physical camera in an offline or realtime renderer.

    Parameters
    ----------
    T
        Exposure value :math:`EV100`.
    T
        Transmission factor of the lens :math:`T`.
    f_v
        Vignetting factor :math:`f_v`.
    theta
        Angle of image point off axis :math:`\\theta`.

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        Photometric exposure in lux-seconds (:math:`lx.s`).

    Notes
    -----
    -   The saturation based speed, :math:`S_{sat}`, of an electronic still
        picture camera is defined as: :math:`S_{sat}=\\cfrac{78}{H_{sat}}`
        where :math:`H_{sat}` is the minimum focal plane exposure, expressed in
        lux-seconds (:math:`lx.s`), that produces the maximum valid (not
        clipped or bloomed) camera output signal. This provides :math:`1/2`
        "stop" of headroom (41% additional headroom) for specular highlights
        above the signal level that would be obtained from a theoretical 100%
        reflectance object in the scene, so that a theoretical 141% reflectance
        object in the scene would produce a focal plane exposure of
        :math:`H_{sat}`.
    -   Scene luminance :math:`L` scaled with the photometric exposure value
        computed by this definition is almost equal to that given by the
        :func:`colour_hdri.saturation_based_speed_focal_plane_exposure`
        definition.

    References
    ----------
    :cite:`ISO2006`, :cite:`Lagarde2014`

    Examples
    --------
    >>> EV100 = exposure_value_100(8, 1 / 250, 400)
    >>> H = photometric_exposure_scale_factor_Lagarde2014(EV100)
    >>> print(H)  # doctest: +ELLIPSIS
    0.0002088...
    >>> H * 4000  # doctest: +ELLIPSIS
    0.8353523...
    """
    EV100 = as_float_array(EV100)

    q = q_factor(T, f_v, theta)

    return as_float(1 / (78 / (100 * q) * 2**EV100))



from .grossberg2003 import samples_Grossberg2003
from .variance_minimization import (
    light_probe_sampling_variance_minimization_Viriyothai2009,
)

__all__ = [
    "samples_Grossberg2003",
]
__all__ += [
    "light_probe_sampling_variance_minimization_Viriyothai2009",
]



"""
EXIF Data Manipulation
======================

EXIF data manipulation routines based on *exiftool*:

-   :func:`colour_hdri.parse_exif_data`
-   :func:`colour_hdri.read_exif_tags`
-   :func:`colour_hdri.copy_exif_tags`
-   :func:`colour_hdri.update_exif_tags`
-   :func:`colour_hdri.delete_exif_tags`
-   :func:`colour_hdri.read_exif_tag`
-   :func:`colour_hdri.write_exif_tag`
"""

from __future__ import annotations

import logging
import numpy as np
import re
import subprocess  # nosec
from collections import defaultdict
from dataclasses import dataclass, field
from fractions import Fraction

from colour.hints import (
    Boolean,
    DTypeFloating,
    DTypeNumber,
    Floating,
    List,
    NDArray,
    Number,
    Optional,
    Sequence,
    SupportsIndex,
    Type,
    Union,
    cast,
)

from colour.constants import DEFAULT_FLOAT_DTYPE
from colour.utilities import as_array, as_float_scalar, optional
from colour.utilities.documentation import (
    DocstringText,
    is_documentation_building,
)

from colour_hdri.utilities import vivification

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "EXIF_EXECUTABLE",
    "EXIFTag",
    "parse_exif_string",
    "parse_exif_number",
    "parse_exif_fraction",
    "parse_exif_array",
    "parse_exif_data",
    "read_exif_tags",
    "copy_exif_tags",
    "update_exif_tags",
    "delete_exif_tags",
    "read_exif_tag",
    "write_exif_tag",
]

EXIF_EXECUTABLE: str = "exiftool"
if is_documentation_building():  # pragma: no cover
    EXIF_EXECUTABLE = DocstringText(EXIF_EXECUTABLE)
    EXIF_EXECUTABLE.__doc__ = """
Command line EXIF manipulation application, usually Phil Harvey's *ExifTool*.
"""


@dataclass
class EXIFTag:
    """
    EXIF tag data.

    Parameters
    ----------
    group
        EXIF tag group name.
    name
        EXIF tag name.
    value
        EXIF tag value.
    identifier
        EXIF tag identifier.
    """

    group: Optional[str] = field(default_factory=lambda: None)
    name: Optional[str] = field(default_factory=lambda: None)
    value: Optional[str] = field(default_factory=lambda: None)
    identifier: Optional[str] = field(default_factory=lambda: None)


def parse_exif_string(exif_tag: EXIFTag) -> str:
    """
    Parse given EXIF tag assuming it is a string and return its value.

    Parameters
    ----------
    exif_tag
        EXIF tag to parse.

    Returns
    -------
    :class:`str`
        Parsed EXIF tag value.
    """

    return str(exif_tag.value)


def parse_exif_number(
    exif_tag: EXIFTag, dtype: Optional[Type[DTypeNumber]] = None
) -> Number:
    """
    Parse given EXIF tag assuming it is a number type and return its value.

    Parameters
    ----------
    exif_tag
        EXIF tag to parse.
    dtype
        Return value data type.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.integer`
        Parsed EXIF tag value.
    """

    dtype = cast(Type[DTypeNumber], optional(dtype, DEFAULT_FLOAT_DTYPE))

    return dtype(exif_tag.value)  # type: ignore[arg-type, return-value]


def parse_exif_fraction(
    exif_tag: EXIFTag, dtype: Optional[Type[DTypeFloating]] = None
) -> Floating:
    """
    Parse given EXIF tag assuming it is a fraction and return its value.

    Parameters
    ----------
    exif_tag
        EXIF tag to parse.
    dtype
        Return value data type.

    Returns
    -------
    :class:`numpy.floating`
        Parsed EXIF tag value.
    """

    dtype = cast(Type[DTypeFloating], optional(dtype, DEFAULT_FLOAT_DTYPE))

    value = (
        exif_tag.value
        if exif_tag.value is None
        else float(Fraction(exif_tag.value))
    )

    return as_float_scalar(value, dtype)  # type: ignore[arg-type]


def parse_exif_array(
    exif_tag: EXIFTag,
    dtype: Optional[Type[DTypeNumber]] = None,
    shape: Optional[Union[SupportsIndex, Sequence[SupportsIndex]]] = None,
) -> NDArray:
    """
    Parse given EXIF tag assuming it is an array and return its value.

    Parameters
    ----------
    exif_tag
        EXIF tag to parse.
    dtype
        Return value data type.
    shape
        Shape of the array to be returned.

    Returns
    -------
    :class:`numpy.ndarray`
        Parsed EXIF tag value.
    """

    dtype = cast(Type[DTypeNumber], optional(dtype, DEFAULT_FLOAT_DTYPE))

    value = (
        exif_tag.value if exif_tag.value is None else exif_tag.value.split()
    )

    return np.reshape(as_array(value, dtype), shape)  # type: ignore[arg-type]


def parse_exif_data(data: str) -> List:
    """
    Parse given EXIF data output from *exiftool*.

    Parameters
    ----------
    data
        EXIF data output.

    Returns
    -------
    :class:`list`
        Parsed EXIF data output.

    Raises
    ------
    ValueError
        If the EXIF data output cannot be parsed.
    """

    search = re.search(
        r"\[(?P<group>\w+)\]\s*(?P<id>(\d+|-))?(?P<tag>.*?):(?P<value>.*$)",
        data,
    )

    if search is not None:
        return [
            group.strip() if group is not None else group
            for group in (
                search.group("group"),
                search.group("id"),
                search.group("tag"),
                search.group("value"),
            )
        ]
    else:
        raise ValueError("The EXIF data output cannot be parsed!")


def read_exif_tags(image: str) -> defaultdict:
    """
    Return given image EXIF image tags.

    Parameters
    ----------
    image
        Image file.

    Returns
    -------
    :class:`defaultdict`
        EXIF tags.
    """

    logging.info(f"Reading '{image}' image EXIF data.")

    exif_tags = vivification()
    lines = str(
        subprocess.check_output(  # nosec
            [EXIF_EXECUTABLE, "-D", "-G", "-a", "-u", "-n", image]
        ),
        "utf-8",
        "ignore",
    ).split("\n")

    for line in lines:
        if not line.strip():
            continue

        group, identifier, tag, value = parse_exif_data(line)

        if not exif_tags[group][tag]:
            exif_tags[group][tag] = []

        exif_tags[group][tag].append(EXIFTag(group, tag, value, identifier))

    return exif_tags


def copy_exif_tags(source: str, target: str) -> Boolean:
    """
    Copy given source image file EXIF tag to given image target.

    Parameters
    ----------
    source
        Source image file.
    target
        Target image file.

    Returns
    -------
    :class:`bool`
        Definition success.
    """

    logging.info(f"Copying '{source}' file EXIF data to '{target}' file.")

    arguments = [EXIF_EXECUTABLE, "-overwrite_original", "-TagsFromFile"]
    arguments += [source, target]
    subprocess.check_output(arguments)  # nosec

    return True


# TODO: Find a better name.
def update_exif_tags(images: Sequence[Sequence[str]]) -> Boolean:
    """
    Update given images pairs EXIF tags.

    Parameters
    ----------
    images
        Image pairs to update the EXIF tags of.

    Returns
    -------
    :class:`bool`
        Definition success.
    """

    success = 1
    for (source, target) in images:
        success *= int(copy_exif_tags(source, target))

    return bool(success)


def delete_exif_tags(image: str) -> Boolean:
    """
    Delete all given image EXIF tags.

    Parameters
    ----------
    image
        Image file to delete the EXIF tags from.

    Returns
    -------
    :class:`bool`
        Definition success.
    """

    logging.info(f"Deleting '{image}' image EXIF tags.")

    subprocess.check_output(  # nosec
        [EXIF_EXECUTABLE, "-overwrite_original", "-all=", image]
    )

    return True


def read_exif_tag(image: str, tag: str) -> str:
    """
    Return given image EXIF tag value.

    Parameters
    ----------
    image : str
        Image file to read the EXIF tag value of.
    tag : str
        Tag to read the value of.

    Returns
    -------
    :class:`str`
        Tag value.
    """

    value = (
        str(
            subprocess.check_output(  # nosec
                [EXIF_EXECUTABLE, f"-{tag}", image]
            ),
            "utf-8",
            "ignore",
        )
        .split(":")
        .pop()
        .strip()
    )

    logging.info(f"Reading '{image}' image '{tag}' EXIF tag value: '{value}'")

    return value


def write_exif_tag(image: str, tag: str, value: str) -> Boolean:
    """
    Set given image EXIF tag value.

    Parameters
    ----------
    image : str
        Image file to set the EXIF tag value of.
    tag : str
        Tag to set the value of.
    value : str
        Value to set.

    Returns
    -------
    :class:`bool`
        Definition success.
    """

    logging.info(
        f"Writing '{image}' image '{tag}' EXIF tag with '{value}' value."
    )

    arguments = [EXIF_EXECUTABLE, "-overwrite_original"]
    arguments += [f"-{tag}={value}", image]
    subprocess.check_output(arguments)  # nosec

    return True



# !/usr/bin/env python
"""Defines the unit tests for the :mod:`colour_hdri.generation.radiance` module."""

from __future__ import annotations

import numpy as np
import os
import unittest

from colour import RGB_COLOURSPACES
from colour.hints import List

from colour_hdri import TESTS_RESOURCES_DIRECTORY
from colour_hdri.generation import image_stack_to_radiance_image
from colour_hdri.calibration import camera_response_functions_Debevec1997
from colour_hdri.utilities import ImageStack, filter_files

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "FROBISHER_001_DIRECTORY",
    "GENERATION_DIRECTORY",
    "JPG_IMAGES",
    "TestRadianceImage",
]

FROBISHER_001_DIRECTORY: str = os.path.join(
    TESTS_RESOURCES_DIRECTORY, "frobisher_001"
)

GENERATION_DIRECTORY: str = os.path.join(
    TESTS_RESOURCES_DIRECTORY, "colour_hdri", "generation"
)

JPG_IMAGES: List[str] = filter_files(FROBISHER_001_DIRECTORY, ("jpg",))


class TestRadianceImage(unittest.TestCase):
    """
    Define :func:`colour_hdri.generation.radiance.\
image_stack_to_radiance_image` definition unit tests methods.
    """

    def test_radiance_image(self):
        """
        Test :func:`colour_hdri.generation.\
radiance.image_stack_to_radiance_image` definition.
        """

        image_stack = ImageStack.from_files(JPG_IMAGES)
        image_stack.data = RGB_COLOURSPACES["sRGB"].cctf_decoding(
            image_stack.data
        )

        # Lower precision for unit tests under *travis-ci*.
        np.testing.assert_allclose(
            image_stack_to_radiance_image(image_stack),
            np.load(
                os.path.join(
                    GENERATION_DIRECTORY, "test_radiance_image_linear.npy"
                )
            ),
            rtol=0.0001,
            atol=0.0001,
        )

        # Lower precision for unit tests under *travis-ci*.
        image_stack = ImageStack.from_files(JPG_IMAGES)
        np.testing.assert_allclose(
            image_stack_to_radiance_image(
                image_stack,
                camera_response_functions=(
                    camera_response_functions_Debevec1997(image_stack)
                ),
            ),
            np.load(
                os.path.join(
                    GENERATION_DIRECTORY, "test_radiance_image_crfs.npy"
                )
            ),
            rtol=0.0001,
            atol=0.0001,
        )


if __name__ == "__main__":
    unittest.main()



"""
Adobe DNG SDK Colour Processing
===============================

Defines various objects implementing *Adobe DNG SDK* colour processing:

-   :func:`colour_hdri.xy_to_camera_neutral`
-   :func:`colour_hdri.camera_neutral_to_xy`
-   :func:`colour_hdri.XYZ_to_camera_space_matrix`
-   :func:`colour_hdri.camera_space_to_XYZ_matrix`

The *Adobe DNG SDK* defines the following tags relevant for the current
implementation:

-   *CalibrationIlluminant1* : The illuminant used for the first set of colour
    calibration tags.
-   *CalibrationIlluminant2* : The illuminant used for an optional second set
    of colour calibration tags.
-   *ColorMatrix1* : *ColorMatrix1* defines a transformation matrix that
    converts XYZ values to reference camera native colour space values, under
    the first calibration illuminant.
-   *ColorMatrix2* : *ColorMatrix2* defines a transformation matrix that
    converts XYZ values to reference camera native colour space values, under
    the second calibration illuminant.
-   *CameraCalibration1* : *CameraCalibration1* defines a calibration matrix
    that transforms reference camera native space values to individual camera
    native space values under the first calibration illuminant. This matrix is
    stored separately from the matrix specified by the *ColorMatrix1* tag to
    allow raw converters to swap in replacement colour matrices based on
    *UniqueCameraModel* tag, while still taking advantage of any per-individual
    camera calibration performed by the camera manufacturer.
-   *CameraCalibration2* : *CameraCalibration2* defines a calibration matrix
    that transforms reference camera native space values to individual camera
    native space values under the second calibration illuminant. This matrix is
    stored separately from the matrix specified by the *ColorMatrix2* tag to
    allow raw converters to swap in replacement colour matrices based on
    *UniqueCameraModel* tag, while still taking advantage of any per-individual
    camera calibration performed by the camera manufacturer.
-   *ReductionMatrix1* : *ReductionMatrix1* defines a dimensionality reduction
    matrix for use as the first stage in converting colour camera native space
    values to XYZ values, under the first calibration illuminant. This tag may
    only be used if *ColorPlanes* is greater than 3.
-   *ReductionMatrix2* : *ReductionMatrix2* defines a dimensionality reduction
    matrix for use as the first stage in converting colour camera native space
    values to XYZ values, under the second calibration illuminant. This tag may
    only be used if *ColorPlanes* is greater than 3.
-   *AnalogBalance* : Normally the stored raw values are not white balanced,
    since any digital white balancing will reduce the dynamic range of the
    final image if the user decides to later adjust the white balance; however,
    if camera hardware is capable of white balancing the colour channels before
    the signal is digitized, it can improve the dynamic range of the final
    image. *AnalogBalance* defines the gain, either analog (recommended) or
    digital (not recommended) that has been applied the stored raw values.
-   *AsShotNeutral* : *AsShotNeutral* specifies the selected white balance at
    time of capture, encoded as the coordinates of a perfectly neutral colour
    in linear reference space values. The inclusion of this tag precludes the
    inclusion of the *AsShotWhiteXY* tag.
-   *AsShotWhiteXY* : *AsShotWhiteXY* specifies the selected white balance at
    time of capture, encoded as x-y chromaticity coordinates. The inclusion of
    this tag precludes the inclusion of the *AsShotNeutral* tag.
-   *ForwardMatrix1* : This tag defines a matrix that maps white balanced
    camera colours to XYZ D50 colours.
-   *ForwardMatrix2* : This tag defines a matrix that maps white balanced
    camera colours to XYZ D50 colours.

Notes
-----
-   At least one of the *ColorMatrix1* or *ColorMatrix2* tags must be included
    in the camera profile, the current implementation expects them to be passed
    as identity matrices if not included.
-   If the *ForwardMatrix1* or *ForwardMatrix2* tags are not included in the
    camera profile, the current implementation expects them to be passed as
    identity matrices.
-   The *ReductionMatrix1* and *ReductionMatrix2* tags are ignored by the
    current implementation which expects cameras with 3 colour planes.
-   *DNG 1.2.0.0* and later supports different companies creating the camera
    calibration tags using different reference cameras. When rendering a *DNG*
    file using a camera profile, it is important to know if the selected camera
    profile was designed using the same reference camera used to create the
    camera calibration tags. If so, then the camera calibration tags should be
    used. If not, then it is preferable to ignore the camera calibration tags
    and use identity matrices instead in order to minimize the worse case
    calibration mismatch error. This matching is done by comparing the
    *CameraCalibrationSignature* tag and the *ProfileCalibrationSignature* tag
    for the selected camera profile. If they match, then use the camera
    calibration tags. If not, then use identity matrices.
-   The Hue/Saturation/Value Mapping Table is ignored by the current
    implementation because deemed unsuitable :cite:`McGuffog2012a`.
-   The various matrices used in this module are extracted from a *Canon EOS 5D
    Mark II* camera.

References
----------
-   :cite:`AdobeSystems2012d` : Adobe Systems. (2012). Translating White
    Balance xy Coordinates to Camera Neutral Coordinates. In Digital Negative
    (DNG) Specification (p. 80).
-   :cite:`AdobeSystems2012e` : Adobe Systems. (2012). Translating Camera
    Neutral Coordinates to White Balance xy Coordinates. In Digital Negative
    (DNG) Specification (pp. 80-81).
-   :cite:`AdobeSystems2012f` : Adobe Systems. (2012). Digital Negative (DNG)
    Specification (pp. 1-101).
-   :cite:`AdobeSystems2012g` : Adobe Systems. (2012). Camera to XYZ (D50)
    Transform. In Digital Negative (DNG) Specification (p. 81).
-   :cite:`AdobeSystems2015d` : Adobe Systems. (2015). Adobe DNG SDK 1.4.
    http://download.adobe.com/pub/adobe/dng/dng_sdk_1_4.zip
-   :cite:`McGuffog2012a` : McGuffog, S. (2012). Hue Twists in DNG Camera
    Profiles. Retrieved October 29, 2016, from
    http://dcptool.sourceforge.net/Hue%20Twists.html
"""

from __future__ import annotations

import numpy as np

from colour.adaptation import matrix_chromatic_adaptation_VonKries
from colour.algebra import (
    is_identity,
    linear_conversion,
    matrix_dot,
    vector_dot,
)
from colour.hints import ArrayLike, Floating, Literal, NDArray, Union
from colour.constants import EPSILON
from colour.models import UCS_to_uv, XYZ_to_UCS, XYZ_to_xy, xy_to_XYZ
from colour.utilities import as_float_array, tstack
from colour.temperature import uv_to_CCT_Robertson1968

from colour_hdri.models import CCS_ILLUMINANT_ADOBEDNG

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "interpolated_matrix",
    "xy_to_camera_neutral",
    "camera_neutral_to_xy",
    "XYZ_to_camera_space_matrix",
    "camera_space_to_XYZ_matrix",
]


def interpolated_matrix(
    CCT: Floating,
    CCT_1: Floating,
    CCT_2: Floating,
    M_1: ArrayLike,
    M_2: ArrayLike,
) -> NDArray:
    """
    Compute the matrix interpolated from :math:`CCT_1` and :math:`CCT_2`
    correlated colour temperatures to respectively :math:`M_T` and :math:`M_R`
    colour matrices using given correlated colour temperature :math:`CCT`
    interpolation value.

    Parameters
    ----------
    CCT
        Correlated colour temperature :math:`CCT`.
    CCT_1
        Correlated colour temperature :math:`CCT_1`.
    CCT_2
        Correlated colour temperature :math:`CCT_2`.
    M_1
        :math:`M_T` colour matrix.
    M_2
        :math:`M_R` colour matrix.

    Returns
    -------
    :class:`numpy.ndarray`
        Interpolated colour matrix :math:`M_i`.

    Notes
    -----
    -   The computation is performed in mired (MIcro REciprocal Degree,
        reciprocal megakelvin) :math:`MK^{-1}`.

    Examples
    --------
    >>> CCT = 5000
    >>> CCT_1 = 2850
    >>> CCT_2 = 6500
    >>> M_T = np.array([
    ...     [0.5309, -0.0229, -0.0336],
    ...     [-0.6241, 1.3265, 0.3337],
    ...     [-0.0817, 0.1215, 0.6664]])
    >>> M_R = np.array([
    ...     [0.4716, 0.0603, -0.0830],
    ...     [-0.7798, 1.5474, 0.2480],
    ...     [-0.1496, 0.1937, 0.6651]])
    >>> interpolated_matrix(CCT, CCT_1, CCT_2, M_T, M_R)  # doctest: +ELLIPSIS
    array([[ 0.4854908...,  0.0408106..., -0.0714282...],
           [-0.7433278...,  1.4956549...,  0.2680749...],
           [-0.1336946...,  0.1767874...,  0.6654045...]])
    """

    if CCT <= CCT_1:
        return as_float_array(M_1)
    elif CCT >= CCT_2:
        return as_float_array(M_2)
    else:
        return linear_conversion(
            1e6 / CCT, (1e6 / CCT_1, 1e6 / CCT_2), tstack([M_1, M_2])
        )


def xy_to_camera_neutral(
    xy: ArrayLike,
    CCT_calibration_illuminant_1: Floating,
    CCT_calibration_illuminant_2: Floating,
    M_color_matrix_1: ArrayLike,
    M_color_matrix_2: ArrayLike,
    M_camera_calibration_1: ArrayLike,
    M_camera_calibration_2: ArrayLike,
    analog_balance: ArrayLike,
) -> NDArray:
    """
    Convert given *xy* white balance chromaticity coordinates to
    *Camera Neutral* coordinates.

    Parameters
    ----------
    xy
        *xy* white balance chromaticity coordinates.
    CCT_calibration_illuminant_1
        Correlated colour temperature of *CalibrationIlluminant1*.
    CCT_calibration_illuminant_2
        Correlated colour temperature of *CalibrationIlluminant2*.
    M_color_matrix_1
        *ColorMatrix1* tag matrix.
    M_color_matrix_2
        *ColorMatrix2* tag matrix.
    M_camera_calibration_1
        *CameraCalibration1* tag matrix.
    M_camera_calibration_2
        *CameraCalibration2* tag matrix.
    analog_balance
        *AnalogBalance* tag vector.

    Returns
    -------
    :class:`numpy.ndarray`
        *Camera Neutral* coordinates.

    References
    ----------
    :cite:`AdobeSystems2012d`, :cite:`AdobeSystems2012f`,
    :cite:`AdobeSystems2015d`, :cite:`McGuffog2012a`

    Examples
    --------
    >>> M_color_matrix_1 = np.array(
    ...     [[0.5309, -0.0229, -0.0336],
    ...      [-0.6241, 1.3265, 0.3337],
    ...      [-0.0817, 0.1215, 0.6664]])
    >>> M_color_matrix_2 = np.array(
    ...     [[0.4716, 0.0603, -0.0830],
    ...      [-0.7798, 1.5474, 0.2480],
    ...      [-0.1496, 0.1937, 0.6651]])
    >>> M_camera_calibration_1 = np.identity(3)
    >>> M_camera_calibration_2 = np.identity(3)
    >>> analog_balance = np.ones(3)
    >>> xy_to_camera_neutral(  # doctest: +ELLIPSIS
    ...     np.array([0.32816244, 0.34698169]),
    ...     2850,
    ...     6500,
    ...     M_color_matrix_1,
    ...     M_color_matrix_2,
    ...     M_camera_calibration_1,
    ...     M_camera_calibration_2,
    ...     analog_balance)
    array([ 0.4130699...,  1...        ,  0.646465...])
    """

    M_XYZ_to_camera = XYZ_to_camera_space_matrix(
        xy,
        CCT_calibration_illuminant_1,
        CCT_calibration_illuminant_2,
        M_color_matrix_1,
        M_color_matrix_2,
        M_camera_calibration_1,
        M_camera_calibration_2,
        analog_balance,
    )

    camera_neutral = vector_dot(M_XYZ_to_camera, xy_to_XYZ(xy))
    camera_neutral /= camera_neutral[1]

    return camera_neutral


def camera_neutral_to_xy(
    camera_neutral: ArrayLike,
    CCT_calibration_illuminant_1: Floating,
    CCT_calibration_illuminant_2: Floating,
    M_color_matrix_1: ArrayLike,
    M_color_matrix_2: ArrayLike,
    M_camera_calibration_1: ArrayLike,
    M_camera_calibration_2: ArrayLike,
    analog_balance: ArrayLike,
    epsilon: Floating = EPSILON,
) -> NDArray:
    """
    Convert given *Camera Neutral* coordinates to *xy* white balance
    chromaticity coordinates.

    Parameters
    ----------
    camera_neutral
        *Camera Neutral* coordinates.
    CCT_calibration_illuminant_1
        Correlated colour temperature of *CalibrationIlluminant1*.
    CCT_calibration_illuminant_2
        Correlated colour temperature of *CalibrationIlluminant2*.
    M_color_matrix_1
        *ColorMatrix1* tag matrix.
    M_color_matrix_2
        *ColorMatrix2* tag matrix.
    M_camera_calibration_1
        *CameraCalibration1* tag matrix.
    M_camera_calibration_2
        *CameraCalibration2* tag matrix.
    analog_balance
        *AnalogBalance* tag vector.
    epsilon
        Threshold value for computation convergence.

    Returns
    -------
    :class:`numpy.ndarray`
        *xy* white balance chromaticity coordinates.

    Raises
    ------
    RuntimeError
        If the given *Camera Neutral* coordinates did not converge to *xy*
        white balance chromaticity coordinates.

    References
    ----------
    :cite:`AdobeSystems2012e`, :cite:`AdobeSystems2012f`,
    :cite:`AdobeSystems2015d`, :cite:`McGuffog2012a`

    Examples
    --------
    >>> M_color_matrix_1 = np.array(
    ...     [[0.5309, -0.0229, -0.0336],
    ...      [-0.6241, 1.3265, 0.3337],
    ...      [-0.0817, 0.1215, 0.6664]])
    >>> M_color_matrix_2 = np.array(
    ...     [[0.4716, 0.0603, -0.0830],
    ...      [-0.7798, 1.5474, 0.2480],
    ...      [-0.1496, 0.1937, 0.6651]])
    >>> M_camera_calibration_1 = np.identity(3)
    >>> M_camera_calibration_2 = np.identity(3)
    >>> analog_balance = np.ones(3)
    >>> camera_neutral_to_xy(  # doctest: +ELLIPSIS
    ...     np.array([0.413070, 1.000000, 0.646465]),
    ...     2850,
    ...     6500,
    ...     M_color_matrix_1,
    ...     M_color_matrix_2,
    ...     M_camera_calibration_1,
    ...     M_camera_calibration_2,
    ...     analog_balance)
    array([ 0.3281624...,  0.3469816...])
    """

    # Initial *xy* chromaticity coordinates guess.
    xy = np.array((1 / 3, 1 / 3))

    while True:
        xy_p = np.copy(xy)
        M_XYZ_to_camera = XYZ_to_camera_space_matrix(
            xy,
            CCT_calibration_illuminant_1,
            CCT_calibration_illuminant_2,
            M_color_matrix_1,
            M_color_matrix_2,
            M_camera_calibration_1,
            M_camera_calibration_2,
            analog_balance,
        )

        XYZ = vector_dot(np.linalg.inv(M_XYZ_to_camera), camera_neutral)
        xy = XYZ_to_xy(XYZ)

        if np.abs(np.sum(xy_p - xy)) <= epsilon:
            return xy

    raise RuntimeError(
        f'"Camera Neutral" coordinates "{xy}" did not converge to "xy" white '
        f"balance chromaticity coordinates!"
    )


def XYZ_to_camera_space_matrix(
    xy: ArrayLike,
    CCT_calibration_illuminant_1: Floating,
    CCT_calibration_illuminant_2: Floating,
    M_color_matrix_1: ArrayLike,
    M_color_matrix_2: ArrayLike,
    M_camera_calibration_1: ArrayLike,
    M_camera_calibration_2: ArrayLike,
    analog_balance: ArrayLike,
) -> NDArray:
    """
    Return the *CIE XYZ* to *Camera Space* matrix for given *xy* white balance
    chromaticity coordinates.

    Parameters
    ----------
    xy
        *xy* white balance chromaticity coordinates.
    CCT_calibration_illuminant_1
        Correlated colour temperature of *CalibrationIlluminant1*.
    CCT_calibration_illuminant_2
        Correlated colour temperature of *CalibrationIlluminant2*.
    M_color_matrix_1
        *ColorMatrix1* tag matrix.
    M_color_matrix_2
        *ColorMatrix2* tag matrix.
    M_camera_calibration_1
        *CameraCalibration1* tag matrix.
    M_camera_calibration_2
        *CameraCalibration2* tag matrix.
    analog_balance
        *AnalogBalance* tag vector.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* to *Camera Space* matrix.

    Notes
    -----
    -   The reference illuminant is D50 as defined per
        :attr:`colour_hdri.models.datasets.dng.CCS_ILLUMINANT_ADOBEDNG`
        attribute.

    References
    ----------
    :cite:`AdobeSystems2012f`, :cite:`AdobeSystems2015d`, :cite:`McGuffog2012a`

    Examples
    --------
    >>> M_color_matrix_1 = np.array(
    ...     [[0.5309, -0.0229, -0.0336],
    ...      [-0.6241, 1.3265, 0.3337],
    ...      [-0.0817, 0.1215, 0.6664]])
    >>> M_color_matrix_2 = np.array(
    ...     [[0.4716, 0.0603, -0.0830],
    ...      [-0.7798, 1.5474, 0.2480],
    ...      [-0.1496, 0.1937, 0.6651]])
    >>> M_camera_calibration_1 = np.identity(3)
    >>> M_camera_calibration_2 = np.identity(3)
    >>> analog_balance = np.ones(3)
    >>> XYZ_to_camera_space_matrix(  # doctest: +ELLIPSIS
    ...     np.array([0.34510414, 0.35162252]),
    ...     2850,
    ...     6500,
    ...     M_color_matrix_1,
    ...     M_color_matrix_2,
    ...     M_camera_calibration_1,
    ...     M_camera_calibration_2,
    ...     analog_balance)
    array([[ 0.4854908...,  0.0408106..., -0.0714282...],
           [-0.7433278...,  1.4956549...,  0.2680749...],
           [-0.1336946...,  0.1767874...,  0.6654045...]])
    """

    M_AB = np.diagflat(analog_balance)

    uv = UCS_to_uv(XYZ_to_UCS(xy_to_XYZ(xy)))
    CCT, _D_uv = uv_to_CCT_Robertson1968(uv)

    if is_identity(M_color_matrix_1) or is_identity(M_color_matrix_2):
        M_CM = (
            M_color_matrix_1
            if is_identity(M_color_matrix_2)
            else M_color_matrix_2
        )
    else:
        M_CM = interpolated_matrix(
            CCT,
            CCT_calibration_illuminant_1,
            CCT_calibration_illuminant_2,
            M_color_matrix_1,
            M_color_matrix_2,
        )

    M_CC = interpolated_matrix(
        CCT,
        CCT_calibration_illuminant_1,
        CCT_calibration_illuminant_2,
        M_camera_calibration_1,
        M_camera_calibration_2,
    )

    M_XYZ_to_camera_space = matrix_dot(matrix_dot(M_AB, M_CC), M_CM)

    return M_XYZ_to_camera_space


def camera_space_to_XYZ_matrix(
    xy: ArrayLike,
    CCT_calibration_illuminant_1: Floating,
    CCT_calibration_illuminant_2: Floating,
    M_color_matrix_1: ArrayLike,
    M_color_matrix_2: ArrayLike,
    M_camera_calibration_1: ArrayLike,
    M_camera_calibration_2: ArrayLike,
    analog_balance: ArrayLike,
    M_forward_matrix_1: ArrayLike,
    M_forward_matrix_2: ArrayLike,
    chromatic_adaptation_transform: Union[
        Literal[
            "Bianco 2010",
            "Bianco PC 2010",
            "Bradford",
            "CAT02 Brill 2008",
            "CAT02",
            "CAT16",
            "CMCCAT2000",
            "CMCCAT97",
            "Fairchild",
            "Sharp",
            "Von Kries",
            "XYZ Scaling",
        ],
        str,
    ] = "Bradford",
) -> NDArray:
    """
    Return the *Camera Space* to *CIE XYZ* matrix for given *xy* white
    balance chromaticity coordinates.

    Parameters
    ----------
    xy
        *xy* white balance chromaticity coordinates.
    CCT_calibration_illuminant_1
        Correlated colour temperature of *CalibrationIlluminant1*.
    CCT_calibration_illuminant_2
        Correlated colour temperature of *CalibrationIlluminant2*.
    M_color_matrix_1
        *ColorMatrix1* tag matrix.
    M_color_matrix_2
        *ColorMatrix2* tag matrix.
    M_camera_calibration_1
        *CameraCalibration1* tag matrix.
    M_camera_calibration_2
        *CameraCalibration2* tag matrix.
    analog_balance
        *AnalogBalance* tag vector.
    M_forward_matrix_1
        *ForwardMatrix1* tag matrix.
    M_forward_matrix_2
        *ForwardMatrix2* tag matrix.
    chromatic_adaptation_transform
        Chromatic adaptation transform.

    Returns
    -------
    :class:`numpy.ndarray`
        *Camera Space* to *CIE XYZ* matrix.

    Notes
    -----
    -   The reference illuminant is D50 as defined per
        :attr:`colour_hdri.models.datasets.dng.CCS_ILLUMINANT_ADOBEDNG`
        attribute.

    References
    ----------
    :cite:`AdobeSystems2012f`, :cite:`AdobeSystems2012g`,
    :cite:`AdobeSystems2015d`, :cite:`McGuffog2012a`

    Examples
    --------
    >>> M_color_matrix_1 = np.array(
    ...     [[0.5309, -0.0229, -0.0336],
    ...      [-0.6241, 1.3265, 0.3337],
    ...      [-0.0817, 0.1215, 0.6664]])
    >>> M_color_matrix_2 = np.array(
    ...     [[0.4716, 0.0603, -0.0830],
    ...      [-0.7798, 1.5474, 0.2480],
    ...      [-0.1496, 0.1937, 0.6651]])
    >>> M_camera_calibration_1 = np.identity(3)
    >>> M_camera_calibration_2 = np.identity(3)
    >>> analog_balance = np.ones(3)
    >>> M_forward_matrix_1 = np.array(
    ...     [[0.8924, -0.1041, 0.1760],
    ...      [0.4351, 0.6621, -0.0972],
    ...      [0.0505, -0.1562, 0.9308]])
    >>> M_forward_matrix_2 = np.array(
    ...     [[0.8924, -0.1041, 0.1760],
    ...      [0.4351, 0.6621, -0.0972],
    ...      [0.0505, -0.1562, 0.9308]])
    >>> camera_space_to_XYZ_matrix(  # doctest: +ELLIPSIS
    ...     np.array([0.32816244, 0.34698169]),
    ...     2850,
    ...     6500,
    ...     M_color_matrix_1,
    ...     M_color_matrix_2,
    ...     M_camera_calibration_1,
    ...     M_camera_calibration_2,
    ...     analog_balance,
    ...     M_forward_matrix_1,
    ...     M_forward_matrix_2)
    array([[ 2.1604087..., -0.1041...    ,  0.2722498...],
           [ 1.0533324...,  0.6621...    , -0.1503561...],
           [ 0.1222553..., -0.1562...    ,  1.4398304...]])
    """

    # *ForwardMatrix1* and *ForwardMatrix2* are not included in the camera
    # profile.
    if is_identity(M_forward_matrix_1) and is_identity(M_forward_matrix_2):
        M_camera_to_XYZ = np.linalg.inv(
            XYZ_to_camera_space_matrix(
                xy,
                CCT_calibration_illuminant_1,
                CCT_calibration_illuminant_2,
                M_color_matrix_1,
                M_color_matrix_2,
                M_camera_calibration_1,
                M_camera_calibration_2,
                analog_balance,
            )
        )
        M_CAT = matrix_chromatic_adaptation_VonKries(
            xy_to_XYZ(xy),
            xy_to_XYZ(CCS_ILLUMINANT_ADOBEDNG),
            chromatic_adaptation_transform,
        )
        M_camera_space_to_XYZ = matrix_dot(M_CAT, M_camera_to_XYZ)
    else:
        uv = UCS_to_uv(XYZ_to_UCS(xy_to_XYZ(xy)))
        CCT, _D_uv = uv_to_CCT_Robertson1968(uv)

        M_CC = interpolated_matrix(
            CCT,
            CCT_calibration_illuminant_1,
            CCT_calibration_illuminant_2,
            M_camera_calibration_1,
            M_camera_calibration_2,
        )

        # The reference implementation :cite:`AdobeSystems2015d` diverges from
        # the white-paper :cite:`AdobeSystems2012f`:
        # The reference implementation directly computes the camera neutral by
        # multiplying directly the interpolated colour matrix :math:`CM` with
        # the tristimulus values of the *xy* white balance chromaticity
        # coordinates.
        # The current implementation is based on the white-paper so that the
        # interpolated camera calibration matrix :math:`CC` and the
        # analog balance matrix :math:`AB` are accounted for.
        camera_neutral = xy_to_camera_neutral(
            xy,
            CCT_calibration_illuminant_1,
            CCT_calibration_illuminant_2,
            M_color_matrix_1,
            M_color_matrix_2,
            M_camera_calibration_1,
            M_camera_calibration_2,
            analog_balance,
        )

        M_AB = np.diagflat(analog_balance)

        M_reference_neutral = vector_dot(
            np.linalg.inv(matrix_dot(M_AB, M_CC)), camera_neutral
        )
        M_D = np.linalg.inv(np.diagflat(M_reference_neutral))
        M_FM = interpolated_matrix(
            CCT,
            CCT_calibration_illuminant_1,
            CCT_calibration_illuminant_2,
            M_forward_matrix_1,
            M_forward_matrix_2,
        )
        M_camera_space_to_XYZ = matrix_dot(
            matrix_dot(M_FM, M_D), np.linalg.inv(matrix_dot(M_AB, M_CC))
        )

    return M_camera_space_to_XYZ



# !/usr/bin/env python
"""
Defines the unit tests for the
:mod:`colour_hdri.sampling.variance_minimization` module.
"""

from __future__ import annotations

import numpy as np
import os
import unittest

from colour import RGB_COLOURSPACES, RGB_luminance, read_image

from colour_hdri import TESTS_RESOURCES_DIRECTORY
from colour_hdri.sampling import (
    light_probe_sampling_variance_minimization_Viriyothai2009,
)
from colour_hdri.sampling.variance_minimization import (
    luminance_variance,
    find_regions_variance_minimization_Viriyothai2009,
    highlight_regions_variance_minimization,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "SAMPLING_DIRECTORY",
    "TestLuminanceVariance",
    "TestFindRegionsVarianceMinimizationViriyothai2009",
    "TestLightProbeSamplingVarianceMinimizationViriyothai2009",
]

SAMPLING_DIRECTORY: str = os.path.join(
    TESTS_RESOURCES_DIRECTORY, "colour_hdri", "sampling"
)


class TestLuminanceVariance(unittest.TestCase):
    """
    Define :func:`colour_hdri.sampling.variance_minimization.\
luminance_variance` definition unit tests methods.
    """

    def test_luminance_variance(self):
        """
        Test :func:`colour_hdri.sampling.variance_minimization.\
luminance_variance` definition.
        """

        a = np.tile(np.arange(5), (5, 1))

        self.assertAlmostEqual(luminance_variance(a), 12.24744871, places=7)


class TestFindRegionsVarianceMinimizationViriyothai2009(unittest.TestCase):
    """
    Define :func:`colour_hdri.sampling.variance_minimization.\
find_regions_variance_minimization_Viriyothai2009` definition unit tests
    methods.
    """

    def test_find_regions_variance_minimization_Viriyothai2009(self):
        """
        Test :func:`colour_hdri.sampling.variance_minimization.\
find_regions_variance_minimization_Viriyothai2009` definition.
        """

        colourspace = RGB_COLOURSPACES["sRGB"]

        image = read_image(
            str(
                os.path.join(
                    SAMPLING_DIRECTORY,
                    "tests_light_probe_sampling_variance_minimization_"
                    "Viriyothai2009.exr",
                )
            )
        )

        Y = RGB_luminance(image, colourspace.primaries, colourspace.whitepoint)

        regions = find_regions_variance_minimization_Viriyothai2009(Y, n=1)
        self.assertListEqual(regions, [(0, 256, 0, 156), (0, 256, 156, 256)])

        regions = find_regions_variance_minimization_Viriyothai2009(Y, n=2)
        self.assertListEqual(
            regions,
            [
                (0, 97, 0, 156),
                (97, 256, 0, 156),
                (0, 100, 156, 256),
                (100, 256, 156, 256),
            ],
        )

        regions = find_regions_variance_minimization_Viriyothai2009(Y, n=4)
        self.assertListEqual(
            regions,
            [
                (0, 39, 0, 91),
                (39, 97, 0, 91),
                (0, 39, 91, 156),
                (39, 97, 91, 156),
                (97, 159, 0, 92),
                (97, 159, 92, 156),
                (159, 256, 0, 93),
                (159, 256, 93, 156),
                (0, 42, 156, 216),
                (42, 100, 156, 216),
                (0, 44, 216, 256),
                (44, 100, 216, 256),
                (100, 163, 156, 215),
                (100, 163, 215, 256),
                (163, 256, 156, 216),
                (163, 256, 216, 256),
            ],
        )


class TestHighlightRegionsVarianceMinimization(unittest.TestCase):
    """
    Define :func:`colour_hdri.sampling.variance_minimization.\
highlight_regions_variance_minimization` definition unit tests methods.
    """

    def test_highlight_regions_variance_minimization(self):
        """
        Test :func:`colour_hdri.sampling.variance_minimization.\
highlight_regions_variance_minimization` definition.
        """

        colourspace = RGB_COLOURSPACES["sRGB"]

        image = read_image(
            str(
                os.path.join(
                    SAMPLING_DIRECTORY,
                    "tests_light_probe_sampling_variance_minimization_"
                    "Viriyothai2009.exr",
                )
            )
        )

        Y = RGB_luminance(image, colourspace.primaries, colourspace.whitepoint)
        regions = find_regions_variance_minimization_Viriyothai2009(Y, n=4)
        np.testing.assert_almost_equal(
            highlight_regions_variance_minimization(image, regions),
            read_image(
                str(
                    os.path.join(
                        SAMPLING_DIRECTORY,
                        "tests_highlight_regions_variance_minimization.exr",
                    )
                )
            ),
            decimal=7,
        )


class TestLightProbeSamplingVarianceMinimizationViriyothai2009(
    unittest.TestCase
):
    """
    Define :func:`colour_hdri.sampling.variance_minimization.\
light_probe_sampling_variance_minimization_Viriyothai2009` definition unit
    tests methods.
    """

    def test_light_probe_sampling_variance_minimization_Viriyothai2009(self):
        """
        Test :func:`colour_hdri.sampling.variance_minimization.\
light_probe_sampling_variance_minimization_Viriyothai2009` definition.
        """

        image = read_image(
            str(
                os.path.join(
                    SAMPLING_DIRECTORY,
                    "tests_light_probe_sampling_variance_minimization_"
                    "Viriyothai2009.exr",
                )
            )
        )

        lights = light_probe_sampling_variance_minimization_Viriyothai2009(
            image
        )
        uvs = np.array([light[0] for light in lights])
        colours = np.array([light[1] for light in lights])
        indexes = np.array([light[2] for light in lights])

        np.testing.assert_almost_equal(
            uvs,
            np.array(
                [
                    [0.16015625, 0.11328125],
                    [0.15625000, 0.32421875],
                    [0.42968750, 0.11328125],
                    [0.42968750, 0.32031250],
                    [0.16796875, 0.48437500],
                    [0.43359375, 0.48437500],
                    [0.18359375, 0.74218750],
                    [0.44140625, 0.75390625],
                    [0.64843750, 0.11718750],
                    [0.64843750, 0.35156250],
                    [0.87500000, 0.11718750],
                    [0.87500000, 0.35937500],
                    [0.64062500, 0.54687500],
                    [0.87500000, 0.54687500],
                    [0.64843750, 0.79687500],
                    [0.87500000, 0.80078125],
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            colours,
            np.array(
                [
                    [40.65992016, 226.11660475, 266.78098774],
                    [30.73776919, 130.37145448, 161.10773420],
                    [98.10281688, 201.29676312, 299.40549910],
                    [74.89445847, 117.00525796, 191.89859456],
                    [42.24291545, 125.58142895, 167.82099468],
                    [90.15780473, 102.82409275, 192.97436064],
                    [82.17253280, 97.55175847, 179.72394174],
                    [152.10083479, 72.06310755, 224.16233468],
                    [128.44782221, 173.29928458, 301.74819988],
                    [105.67531514, 104.05130512, 209.73328489],
                    [196.73262107, 196.16986090, 392.89716852],
                    [154.23001331, 111.56532115, 265.79564852],
                    [120.04539376, 81.22123903, 201.26542896],
                    [191.57947493, 95.21154106, 286.79548484],
                    [168.29435712, 45.09299320, 213.38641733],
                    [253.65272349, 50.30476046, 303.96245855],
                ]
            ),
            decimal=7,
        )

        np.testing.assert_array_equal(
            indexes,
            np.array(
                [
                    [29, 41],
                    [83, 40],
                    [29, 110],
                    [82, 110],
                    [124, 43],
                    [124, 111],
                    [190, 47],
                    [193, 113],
                    [30, 166],
                    [90, 166],
                    [30, 224],
                    [92, 224],
                    [140, 164],
                    [140, 224],
                    [204, 166],
                    [205, 224],
                ]
            ),
        )


if __name__ == "__main__":
    unittest.main()



# !/usr/bin/env python
"""Defines the unit tests for the :mod:`colour_hdri.models.rgb` module."""

import numpy as np
import unittest

from colour_hdri.models import camera_space_to_RGB, camera_space_to_sRGB

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestCameraSpaceToRGB",
    "TestCameraSpaceTosRGB",
]


class TestCameraSpaceToRGB(unittest.TestCase):
    """
    Define :func:`colour_hdri.models.rgb.camera_space_to_RGB` definition
    unit tests methods.
    """

    def test_camera_space_to_RGB(self):
        """Test :func:`colour_hdri.models.rgb.camera_space_to_RGB` definition."""

        np.testing.assert_almost_equal(
            camera_space_to_RGB(
                np.array([0.80660, 0.81638, 0.65885]),
                np.array(
                    [
                        [0.47160000, 0.06030000, -0.08300000],
                        [-0.77980000, 1.54740000, 0.24800000],
                        [-0.14960000, 0.19370000, 0.66510000],
                    ]
                ),
                np.array(
                    [
                        [0.41238656, 0.35759149, 0.18045049],
                        [0.21263682, 0.71518298, 0.07218020],
                        [0.01933062, 0.11919716, 0.95037259],
                    ]
                ),
            ),
            np.array([0.75641808, 0.86831924, 0.60445896]),
            decimal=7,
        )


class TestCameraSpaceTosRGB(unittest.TestCase):
    """
    Define :func:`colour_hdri.models.rgb.camera_space_to_sRGB` definition
    unit tests methods.
    """

    def test_camera_space_to_sRGB(self):
        """Test :func:`colour_hdri.models.rgb.camera_space_to_sRGB` definition."""

        np.testing.assert_almost_equal(
            camera_space_to_sRGB(
                np.array([0.80660, 0.81638, 0.65885]),
                np.array(
                    [
                        [0.47160000, 0.06030000, -0.08300000],
                        [-0.77980000, 1.54740000, 0.24800000],
                        [-0.14960000, 0.19370000, 0.66510000],
                    ]
                ),
            ),
            np.array([0.75643502, 0.86831555, 0.60447061]),
            decimal=7,
        )


if __name__ == "__main__":
    unittest.main()



"""
Image Data & Metadata Utilities
===============================

Defines various image data and metadata utilities classes:

-   :class:`colour_hdri.Metadata`
-   :class:`colour_hdri.Image`
-   :class:`colour_hdri.ImageStack`
"""

from __future__ import annotations

import logging
import numpy as np
from collections.abc import MutableSequence
from dataclasses import dataclass, field, fields

from colour.hints import (
    Any,
    ArrayLike,
    Callable,
    FloatingOrNDArray,
    Integer,
    List,
    NDArray,
    Optional,
    Sequence,
    Union,
    cast,
)
from colour import read_image
from colour.utilities import (
    MixinDataclassArray,
    as_float_array,
    is_string,
    tsplit,
    tstack,
    warning,
)

from colour_hdri.exposure import average_luminance
from colour_hdri.utilities.exif import (
    parse_exif_array,
    parse_exif_fraction,
    parse_exif_number,
    read_exif_tags,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "Metadata",
    "Image",
    "ImageStack",
]


@dataclass
class Metadata(MixinDataclassArray):
    """
    Define the base object for storing exif metadata relevant to
    HDRI / radiance image generation.

    Parameters
    ----------
    f_number
        Image *FNumber*.
    exposure_time
        Image *Exposure Time*.
    iso
        Image *ISO*.
    black_level
        Image *Black Level*.
    white_level
        Image *White Level*.
    white_balance_multipliers
        Image white balance multipliers, usually the *As Shot Neutral*  matrix.
    """

    f_number: Optional[NDArray] = field(default_factory=lambda: None)
    exposure_time: Optional[NDArray] = field(default_factory=lambda: None)
    iso: Optional[NDArray] = field(default_factory=lambda: None)
    black_level: Optional[NDArray] = field(default_factory=lambda: None)
    white_level: Optional[NDArray] = field(default_factory=lambda: None)
    white_balance_multipliers: Optional[NDArray] = field(
        default_factory=lambda: None
    )


class Image:
    """
    Define the base object for storing an image along its path, pixel data and
    metadata needed for HDRI / radiance images generation.

    Parameters
    ----------
    path
        Image path.
    data
        Image pixel data array.
    metadata
        Image exif metadata.

    Attributes
    ----------
    -   :attr:`colour_hdri.Image.path`
    -   :attr:`colour_hdri.Image.data`
    -   :attr:`colour_hdri.Image.metadata`

    Methods
    -------
    -   :meth:`colour_hdri.Image.__init__`
    -   :meth:`colour_hdri.Image.read_data`
    -   :meth:`colour_hdri.Image.read_metadata`
    """

    def __init__(
        self,
        path: Optional[str] = None,
        data: Optional[ArrayLike] = None,
        metadata: Optional[Metadata] = None,
    ):
        self._path: Optional[str] = None
        self.path = path
        # TODO: Remove pragma when https://github.com/python/mypy/issues/3004
        # is resolved.
        self._data: Optional[NDArray] = None
        self.data = data  # type: ignore[assignment]
        self._metadata: Optional[Metadata] = None
        self.metadata = metadata

    @property
    def path(self) -> Optional[str]:
        """
        Getter and setter property for the image path.

        Parameters
        ----------
        value
            Value to set the image path with.

        Returns
        -------
        :py:data:`None` or :class:`str`
            Image path.
        """

        return self._path

    @path.setter
    def path(self, value: Optional[str]):
        """Setter for the **self._path** property."""

        if value is not None:
            assert is_string(
                value
            ), f'"path" property: "{value}" type is not "str"!'

        self._path = value

    @property
    def data(self) -> Optional[NDArray]:
        """
        Getter and setter property for the image data.

        Parameters
        ----------
        value
            Value to set the image data with.

        Returns
        -------
        :py:data:`None` or :class:`numpy.ndarray`
            Image data.
        """

        return self._data

    @data.setter
    def data(self, value: Optional[ArrayLike]):
        """Setter for the **self._data** property."""

        if value is not None:
            assert isinstance(value, (tuple, list, np.ndarray, np.matrix)), (
                f'"data" property: "{value}" is not a "tuple", "list", "ndarray" '
                'or "matrix" instance!'
            )

            value = as_float_array(value)

        self._data = value

    @property
    def metadata(self) -> Optional[Metadata]:
        """
        Getter and setter property for the image metadata.

        Parameters
        ----------
        value
            Value to set the image metadata with.

        Returns
        -------
        :py:data:`None` or :class:`colour_hdri.Metadata`
            Image metadata.
        """

        return self._metadata

    @metadata.setter
    def metadata(self, value):
        """Setter for the **self._metadata** property."""

        if value is not None:
            assert isinstance(
                value, Metadata
            ), f'"metadata" property: "{value}" is not a "Metadata" instance!'

        self._metadata = value

    def read_data(self, cctf_decoding: Optional[Callable] = None) -> NDArray:
        """
        Read image pixel data at :attr:`Image.path` attribute.

        Parameters
        ----------
        cctf_decoding
            Decoding colour component transfer function (Decoding CCTF) or
            electro-optical transfer function (EOTF / EOCF).

        Returns
        -------
        :class:`numpy.ndarray`
            Image pixel data.

        Raises
        ------
        ValueError
            If the image path is undefined.
        """

        if self._path is not None:
            logging.info(f'Reading "{self._path}" image.')

            data = read_image(str(self._path))
            if cctf_decoding is not None:
                data = cctf_decoding(data)

            self.data = data

            return data
        else:
            raise ValueError('The image "path" is undefined!')

    def read_metadata(self) -> Metadata:
        """
        Read image relevant exif metadata at :attr:`Image.path` attribute.

        Returns
        -------
        :class:`colour_hdri.Metadata`
            Image relevant exif metadata.

        Raises
        ------
        ValueError
            If the image path is undefined.
        """

        if self._path is not None:
            logging.info(f'Reading "{self._path}" image metadata.')

            exif_data = read_exif_tags(self._path)

            if not exif_data.get("EXIF"):
                warning(
                    f'"{self._path}" file has no "Exif" data, metadata will '
                    f"be undefined!"
                )
                self.metadata = Metadata(*[None] * 6)
                return self.metadata

            f_number = exif_data["EXIF"].get("F Number")
            if f_number is not None:
                f_number = parse_exif_number(f_number[0])

            exposure_time = exif_data["EXIF"].get("Exposure Time")
            if exposure_time is not None:
                exposure_time = parse_exif_fraction(exposure_time[0])

            iso = exif_data["EXIF"].get("ISO")
            if iso is not None:
                iso = parse_exif_number(iso[0])

            black_level = exif_data["EXIF"].get("Black Level")
            if black_level is not None:
                black_level = parse_exif_array(black_level[0])
                black_level = as_float_array(black_level) / 65535

            white_level = exif_data["EXIF"].get("White Level")
            if white_level is not None:
                white_level = parse_exif_array(white_level[0])
                white_level = as_float_array(white_level) / 65535

            white_balance_multipliers = exif_data["EXIF"].get(
                "As Shot Neutral"
            )
            if white_balance_multipliers is not None:
                white_balance_multipliers = parse_exif_array(
                    white_balance_multipliers[0]
                )
                white_balance_multipliers = (
                    as_float_array(white_balance_multipliers)
                    / white_balance_multipliers[1]
                )

            metadata = Metadata(
                f_number,
                exposure_time,
                iso,
                black_level,
                white_level,
                white_balance_multipliers,
            )

            self._metadata = metadata

            return metadata
        else:
            raise ValueError('The image "path" is undefined!')


class ImageStack(MutableSequence):
    """
    Define a convenient stack storing a sequence of images for HDRI / radiance
    images generation.

    Methods
    -------
    -   :meth:`colour_hdri.ImageStack.__init__`
    -   :meth:`colour_hdri.ImageStack.__getitem__`
    -   :meth:`colour_hdri.ImageStack.__setitem__`
    -   :meth:`colour_hdri.ImageStack.__delitem__`
    -   :meth:`colour_hdri.ImageStack.__len__`
    -   :meth:`colour_hdri.ImageStack.__getattr__`
    -   :meth:`colour_hdri.ImageStack.__setattr__`
    -   :meth:`colour_hdri.ImageStack.sort`
    -   :meth:`colour_hdri.ImageStack.insert`
    -   :meth:`colour_hdri.ImageStack.from_files`
    """

    def __init__(self):
        self._data: List = []

    def __getitem__(
        self, index: Union[Integer, slice]
    ) -> Union[Any, MutableSequence[Any]]:
        """
        Return the :class:`colour_hdri.Image` class instance at given index.

        Parameters
        ----------
        index
            :class:`colour_hdri.Image` class instance.

        Returns
        -------
        :class:`colour_hdri.Image`
            :class:`colour_hdri.Image` class instance at given index.
        """

        return self._data[index]

    def __setitem__(self, index: Union[Integer, slice], value: Any):
        """
        Set given :class:`colour_hdri.Image` class instance at given index.

        Parameters
        ----------
        index
            :class:`colour_hdri.Image` class instance index.
        value
            :class:`colour_hdri.Image` class instance to set.
        """

        self._data[index] = value

    def __delitem__(self, index: Union[Integer, slice]):
        """
        Delete the :class:`colour_hdri.Image` class instance at given index.

        Parameters
        ----------
        index
            :class:`colour_hdri.Image` class instance index.
        """

        del self._data[index]

    def __len__(self) -> Integer:
        """
        Return the :class:`colour_hdri.Image` class instances count.

        Returns
        -------
        :class:`numpy.integer`
            :class:`colour_hdri.Image` class instances count.
        """

        return len(self._data)

    def __getattr__(self, attribute: str) -> Any:
        """
        Return the value from the attribute with given name.

        Parameters
        ----------
        name
            Name of the attribute to get the value from.

        Returns
        -------
        :class:`object`
        """

        try:
            return self.__dict__[attribute]
        except KeyError:
            if hasattr(Image, attribute):
                value = [getattr(image, attribute) for image in self]
                if attribute == "data":
                    return tstack(value)
                else:
                    return tuple(value)
            # TODO: Revise then "MixinDataclassArray" is improved.
            elif attribute in [field.name for field in fields(Metadata)]:
                value = [getattr(image.metadata, attribute) for image in self]
                return as_float_array(value)
            else:
                raise AttributeError(
                    f"'{self.__class__.__name__}' object has no attribute "
                    f"'{attribute}'"
                )

    def __setattr__(self, attribute: str, value: Any):
        """
        Set given value to the attribute with given name.

        Parameters
        ----------
        attribute
            Attribute to set the value of.
        value
            Value to set the attribute with.
        """

        if hasattr(Image, attribute):
            if attribute == "data":
                data = tsplit(value)
                for i, image in enumerate(self):
                    image.data = data[i]
            else:
                for i, image in enumerate(self):
                    setattr(image, attribute, value[i])
        elif attribute in [field.name for field in fields(Metadata)]:
            for i, image in enumerate(self):
                setattr(image.metadata, attribute, value[i])
        else:
            super().__setattr__(attribute, value)

    def insert(self, index: Integer, value: Any):
        """
        Insert given :class:`colour_hdri.Image` class instance at given index.

        Parameters
        ----------
        index
            :class:`colour_hdri.Image` class instance index.
        value
            :class:`colour_hdri.Image` class instance to set.
        """

        self._data.insert(index, value)

    def sort(self, key: Optional[Callable] = None):
        """
        Sort the underlying data structure.

        Parameters
        ----------
        key
            Function of one argument that is used to extract a comparison key
            from each data structure.
        """

        self._data = sorted(self._data, key=key)

    @staticmethod
    def from_files(
        image_files: Sequence[str], cctf_decoding: Optional[Callable] = None
    ) -> ImageStack:
        """
        Return a :class:`colour_hdri.ImageStack` instance from given image
        files.

        Parameters
        ----------
        image_files
            Image files.
        cctf_decoding
            Decoding colour component transfer function (Decoding CCTF) or
            electro-optical transfer function (EOTF / EOCF).

        Returns
        -------
        :class:`colour_hdri.ImageStack`
        """

        image_stack = ImageStack()
        for image_file in image_files:
            image = Image(image_file)
            image.read_data(cctf_decoding)
            image.read_metadata()
            image_stack.append(image)

        def luminance_average_key(image: Image) -> Optional[FloatingOrNDArray]:
            """Comparison key function."""

            metadata = cast(Metadata, image.metadata)

            f_number = metadata.f_number
            exposure_time = metadata.exposure_time
            iso = metadata.iso

            if f_number is None or exposure_time is None or iso is None:
                warning(
                    f'"{image.path}" exposure data is missing, average '
                    f"luminance sorting is inapplicable!"
                )
                return None
            else:
                return 1 / average_luminance(f_number, exposure_time, iso)

        image_stack.sort(luminance_average_key)

        return image_stack



"""
Colour - HDRI
=============

HDRI - Radiance image processing algorithms for *Python*.

Subpackages
-----------
-   calibration: Camera calibration computations.
-   exposure: Exposure computations.
-   examples: Examples for the sub-packages.
-   generation: HDRI / radiance image generation.
-   models: Colour models conversion.
-   plotting: Diagrams, figures, etc...
-   process: Image conversion helpers.
-   recovery: Clipped highlights recovery.
-   resources: Resources sub-modules.
-   sampling: Image sampling routines.
-   tonemapping: Tonemapping operators.
-   utilities: Various utilities and data structures.
"""

import numpy as np
import os
import subprocess  # nosec

import colour

from .utilities import (
    EXIF_EXECUTABLE,
    EXIFTag,
    Image,
    ImageStack,
    Metadata,
    copy_exif_tags,
    delete_exif_tags,
    filter_files,
    parse_exif_array,
    parse_exif_data,
    parse_exif_fraction,
    parse_exif_number,
    parse_exif_string,
    path_exists,
    read_exif_tag,
    read_exif_tags,
    update_exif_tags,
    vivification,
    vivified_to_dict,
    write_exif_tag,
)
from .sampling import (
    light_probe_sampling_variance_minimization_Viriyothai2009,
    samples_Grossberg2003,
)
from .exposure import (
    adjust_exposure,
    arithmetic_mean_focal_plane_exposure,
    average_illuminance,
    average_luminance,
    exposure_index_values,
    exposure_value_100,
    photometric_exposure_scale_factor_Lagarde2014,
    focal_plane_exposure,
    illuminance_to_exposure_value,
    luminance_to_exposure_value,
    saturation_based_speed_focal_plane_exposure,
)
from .generation import (
    normal_distribution_function,
    hat_function,
    weighting_function_Debevec1997,
    image_stack_to_radiance_image,
)
from .calibration import (
    absolute_luminance_calibration_Lagarde2016,
    camera_response_functions_Debevec1997,
    g_solve,
    upper_hemisphere_illuminance_weights_Lagarde2016,
)
from .models import (
    camera_neutral_to_xy,
    camera_space_to_RGB,
    camera_space_to_sRGB,
    camera_space_to_XYZ_matrix,
    xy_to_camera_neutral,
    XYZ_to_camera_space_matrix,
)
from .process import (
    DNG_CONVERSION_ARGUMENTS,
    DNG_CONVERTER,
    DNG_EXIF_TAGS_BINDING,
    RAW_CONVERSION_ARGUMENTS,
    RAW_CONVERTER,
    RAW_D_CONVERSION_ARGUMENTS,
    convert_dng_files_to_intermediate_files,
    convert_raw_files_to_dng_files,
    read_dng_files_exif_tags,
)
from .recovery import highlights_recovery_blend, highlights_recovery_LCHab
from .tonemapping import (
    tonemapping_operator_exponential,
    tonemapping_operator_exponentiation_mapping,
    tonemapping_operator_filmic,
    tonemapping_operator_gamma,
    tonemapping_operator_logarithmic,
    tonemapping_operator_logarithmic_mapping,
    tonemapping_operator_normalisation,
    tonemapping_operator_Reinhard2004,
    tonemapping_operator_Schlick1994,
    tonemapping_operator_simple,
    tonemapping_operator_Tumblin1999,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "EXIF_EXECUTABLE",
    "EXIFTag",
    "Image",
    "ImageStack",
    "Metadata",
    "copy_exif_tags",
    "delete_exif_tags",
    "filter_files",
    "parse_exif_array",
    "parse_exif_data",
    "parse_exif_fraction",
    "parse_exif_number",
    "parse_exif_string",
    "path_exists",
    "read_exif_tag",
    "read_exif_tags",
    "update_exif_tags",
    "vivification",
    "vivified_to_dict",
    "write_exif_tag",
]
__all__ += [
    "light_probe_sampling_variance_minimization_Viriyothai2009",
    "samples_Grossberg2003",
]
__all__ += [
    "adjust_exposure",
    "arithmetic_mean_focal_plane_exposure",
    "average_illuminance",
    "average_luminance",
    "exposure_index_values",
    "exposure_value_100",
    "photometric_exposure_scale_factor_Lagarde2014",
    "focal_plane_exposure",
    "illuminance_to_exposure_value",
    "luminance_to_exposure_value",
    "saturation_based_speed_focal_plane_exposure",
]
__all__ += [
    "normal_distribution_function",
    "hat_function",
    "weighting_function_Debevec1997",
    "image_stack_to_radiance_image",
]
__all__ += [
    "absolute_luminance_calibration_Lagarde2016",
    "camera_response_functions_Debevec1997",
    "g_solve",
    "upper_hemisphere_illuminance_weights_Lagarde2016",
]
__all__ += [
    "camera_neutral_to_xy",
    "camera_space_to_RGB",
    "camera_space_to_sRGB",
    "camera_space_to_XYZ_matrix",
    "xy_to_camera_neutral",
    "XYZ_to_camera_space_matrix",
]
__all__ += [
    "DNG_CONVERSION_ARGUMENTS",
    "DNG_CONVERTER",
    "DNG_EXIF_TAGS_BINDING",
    "RAW_CONVERSION_ARGUMENTS",
    "RAW_CONVERTER",
    "RAW_D_CONVERSION_ARGUMENTS",
    "convert_dng_files_to_intermediate_files",
    "convert_raw_files_to_dng_files",
    "read_dng_files_exif_tags",
]
__all__ += [
    "highlights_recovery_blend",
    "highlights_recovery_LCHab",
]
__all__ += [
    "tonemapping_operator_exponential",
    "tonemapping_operator_exponentiation_mapping",
    "tonemapping_operator_filmic",
    "tonemapping_operator_gamma",
    "tonemapping_operator_logarithmic",
    "tonemapping_operator_logarithmic_mapping",
    "tonemapping_operator_normalisation",
    "tonemapping_operator_Reinhard2004",
    "tonemapping_operator_Schlick1994",
    "tonemapping_operator_simple",
    "tonemapping_operator_Tumblin1999",
]

RESOURCES_DIRECTORY = os.path.join(os.path.dirname(__file__), "resources")
EXAMPLES_RESOURCES_DIRECTORY = os.path.join(
    RESOURCES_DIRECTORY, "colour-hdri-examples-datasets"
)
TESTS_RESOURCES_DIRECTORY = os.path.join(
    RESOURCES_DIRECTORY, "colour-hdri-tests-datasets"
)

__application_name__ = "Colour - HDRI"

__major_version__ = "0"
__minor_version__ = "2"
__change_version__ = "0"
__version__ = ".".join(
    (__major_version__, __minor_version__, __change_version__)
)

try:
    _version: str = (
        subprocess.check_output(  # nosec
            ["git", "describe"],
            cwd=os.path.dirname(__file__),
            stderr=subprocess.STDOUT,
        )
        .strip()
        .decode("utf-8")
    )
except Exception:
    _version: str = __version__  # type: ignore[no-redef]

colour.utilities.ANCILLARY_COLOUR_SCIENCE_PACKAGES["colour-hdri"] = _version

del _version

# TODO: Remove legacy printing support when deemed appropriate.
try:
    np.set_printoptions(legacy="1.13")
except TypeError:
    pass



# !/usr/bin/env python
"""Defines the unit tests for the :mod:`colour_hdri.exposure.common` module."""

import numpy as np
import unittest

from colour_hdri.exposure import (
    average_luminance,
    average_illuminance,
    luminance_to_exposure_value,
    illuminance_to_exposure_value,
    adjust_exposure,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestAverageLuminance",
    "TestAverageIlluminance",
    "TestLuminanceToExposureValue",
    "TestAdjustExposure",
]


class TestAverageLuminance(unittest.TestCase):
    """
    Define :func:`colour_hdri.exposure.common.average_luminance` definition
    unit tests methods.
    """

    def test_average_luminance(self):
        """Test :func:`colour_hdri.exposure.common.average_luminance` definition."""

        np.testing.assert_almost_equal(
            average_luminance(
                np.array([2.8, 5.6, 8]),
                np.array([0.125, 0.5, 1.0]),
                np.array([100, 800, 16000]),
            ),
            np.array([7.84000000, 0.98000000, 0.05000000]),
            decimal=7,
        )


class TestAverageIlluminance(unittest.TestCase):
    """
    Define :func:`colour_hdri.exposure.common.average_illuminance` definition
    unit tests methods.
    """

    def test_average_illuminance(self):
        """
        Test :func:`colour_hdri.exposure.common.average_illuminance`
        definition.
        """

        np.testing.assert_almost_equal(
            average_illuminance(
                np.array([2.8, 5.6, 8]),
                np.array([0.125, 0.5, 1.0]),
                np.array([100, 800, 16000]),
            ),
            np.array([156.80000000, 19.60000000, 1.00000000]),
            decimal=7,
        )


class TestLuminanceToExposureValue(unittest.TestCase):
    """
    Define :func:`colour_hdri.exposure.common.luminance_to_exposure_value`
    definition unit tests methods.
    """

    def test_luminance_to_exposure_value(self):
        """
        Test :func:`colour_hdri.exposure.common.luminance_to_exposure_value`
        definition.
        """

        np.testing.assert_almost_equal(
            luminance_to_exposure_value(
                np.array([0.125, 0.250, 0.125]),
                np.array([100, 100, 100]),
                np.array([12.5, 12.5, 14]),
            ),
            np.array([0.00000000, 1.00000000, -0.16349873]),
            decimal=7,
        )


class TestIlluminanceToExposureValue(unittest.TestCase):
    """
    Define :func:`colour_hdri.exposure.common.illuminance_to_exposure_value`
    definition unit tests methods.
    """

    def test_illuminance_to_exposure_value(self):
        """
        Test :func:`colour_hdri.exposure.common.illuminance_to_exposure_value`
        definition.
        """

        np.testing.assert_almost_equal(
            illuminance_to_exposure_value(
                np.array([2.5, 5.0, 0.125]),
                np.array([100, 100, 100]),
                np.array([250, 250, 340]),
            ),
            np.array([0.00000000, 1.00000000, -4.76553475]),
            decimal=7,
        )


class TestAdjustExposure(unittest.TestCase):
    """
    Define :func:`colour_hdri.exposure.common.adjust_exposure` definition
    unit tests methods.
    """

    def test_adjust_exposure(self):
        """Test :func:`colour_hdri.exposure.common.adjust_exposure` definition."""

        np.testing.assert_almost_equal(
            adjust_exposure(np.array([0.25, 0.5, 0.75, 1]), 1),
            np.array([0.5, 1.0, 1.5, 2.0]),
            decimal=7,
        )


if __name__ == "__main__":
    unittest.main()



"""
Clipped Highlights Recovery
===========================

Defines the clipped highlights recovery objects:

-   :func:`colour_hdri.highlights_recovery_blend`
-   :func:`colour_hdri.highlights_recovery_LCHab`

See Also
--------
`Colour - HDRI - Examples: Merge from Raw Files Jupyter Notebook
<https://github.com/colour-science/colour-hdri/\
blob/master/colour_hdri/examples/examples_merge_from_raw_files.ipynb>`__

References
----------
-   :cite:`Coffin2015a` : Coffin, D. (2015). dcraw.
    https://www.cybercom.net/~dcoffin/dcraw/
"""

from __future__ import annotations

import numpy as np

from colour.algebra import vector_dot
from colour.hints import ArrayLike, Floating, NDArray, Optional
from colour.models import (
    RGB_Colourspace,
    LCHab_to_Lab,
    Lab_to_LCHab,
    Lab_to_XYZ,
    RGB_to_XYZ,
    XYZ_to_Lab,
    XYZ_to_RGB,
    RGB_COLOURSPACE_sRGB,
)
from colour.utilities import tsplit, tstack

__author__ = "Colour Developers"
__copyright__ = "Copyright 2015 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "highlights_recovery_blend",
    "highlights_recovery_LCHab",
]


def highlights_recovery_blend(
    RGB: ArrayLike, multipliers: ArrayLike, threshold: Floating = 0.99
) -> NDArray:
    """
    Perform highlights recovery using *Coffin (1997)* method from *dcraw*.

    Parameters
    ----------
    RGB
        *RGB* colourspace array.
    multipliers
        Normalised camera white level or white balance multipliers.
    threshold
        Threshold for highlights selection.

    Returns
    -------
    :class:`numpy.ndarray`
         Highlights recovered *RGB* colourspace array.

    References
    ----------
    :cite:`Coffin2015a`
    """

    M = np.array(
        [
            [1.0000000, 1.0000000, 1.0000000],
            [1.7320508, -1.7320508, 0.0000000],
            [-1.0000000, -1.0000000, 2.0000000],
        ]
    )

    clipping_level = np.min(multipliers) * threshold

    Lab = vector_dot(M, RGB)

    Lab_c = vector_dot(M, np.minimum(RGB, clipping_level))

    s = np.sum((Lab * Lab)[..., 1:3], axis=2)
    s_c = np.sum((Lab_c * Lab_c)[..., 1:3], axis=2)

    ratio = np.sqrt(s_c / s)
    ratio[np.logical_or(np.isnan(ratio), np.isinf(ratio))] = 1

    Lab[:, :, 1:3] *= np.rollaxis(ratio[np.newaxis], 0, 3)

    RGB_o = vector_dot(np.linalg.inv(M), Lab)

    return RGB_o


def highlights_recovery_LCHab(
    RGB: ArrayLike,
    threshold: Optional[Floating] = None,
    RGB_colourspace: RGB_Colourspace = RGB_COLOURSPACE_sRGB,
) -> NDArray:
    """
    Perform highlights recovery in *CIE L\\*C\\*Hab* colourspace.

    Parameters
    ----------
    RGB
        *RGB* colourspace array.
    threshold
        Threshold for highlights selection, automatically computed
        if not given.
    RGB_colourspace
        Working *RGB* colourspace to perform the *CIE L\\*C\\*Hab* to and from.

    Returns
    -------
    :class:`numpy.ndarray`
         Highlights recovered *RGB* colourspace array.
    """

    L, _C, H = tsplit(
        Lab_to_LCHab(
            XYZ_to_Lab(
                RGB_to_XYZ(
                    RGB,
                    RGB_colourspace.whitepoint,
                    RGB_colourspace.whitepoint,
                    RGB_colourspace.matrix_RGB_to_XYZ,
                ),
                RGB_colourspace.whitepoint,
            )
        )
    )

    _L_c, C_c, _H_c = tsplit(
        Lab_to_LCHab(
            XYZ_to_Lab(
                RGB_to_XYZ(
                    np.clip(RGB, 0, threshold),
                    RGB_colourspace.whitepoint,
                    RGB_colourspace.whitepoint,
                    RGB_colourspace.matrix_RGB_to_XYZ,
                ),
                RGB_colourspace.whitepoint,
            )
        )
    )

    return XYZ_to_RGB(
        Lab_to_XYZ(
            LCHab_to_Lab(tstack([L, C_c, H])), RGB_colourspace.whitepoint
        ),
        RGB_colourspace.whitepoint,
        RGB_colourspace.whitepoint,
        RGB_colourspace.matrix_XYZ_to_RGB,
    )
